#ifdef WIN32
#pragma comment(lib, "c:\\TopfieldDisk\\DataFiles\\SDK\\TAP_SDK_DLL")
#include "TAP_Emulations.h"
#endif

#include "tap.h"
#include "Config.h"
#include "skin.h"
#include "Graphics.h"
#include "tapapifix.h"

#ifndef WIN32
// bdb code
#include "exTap.h"
#include "exTras.h"
#endif

#define ID_RBACK 0x8000adcd  

TAP_ID( ID_RBACK );
TAP_PROGRAM_NAME("Super PiP 1.1b");
TAP_AUTHOR_NAME("Mark");
TAP_DESCRIPTION("Supercharged PiP functionality");
TAP_ETCINFO(__DATE__);

#include "TSRCommander.h"

void setupGraphics();
void drawRoundedRectangle(word x, word y, word w, word h, word border, word border_colour, word fill_colour);

// use for Media Manager support
#define  MM_Request     0x30001
#define  MM_Response    0x30002
#define  MM_Activate    0x30003
#define  MM_Exit		0x30004
#define  Super_PiP      0x30005

#define SPLIT_W (720-30)/2
#define SPLIT_H (576)/2
//10, 147, 360, 288, 360, 147, 360, 288

bool debug = 0;

bool active = FALSE;
static bool first_activation = TRUE, handle_initialisation, edit_config_keys, reverse_pip, pip_generated = FALSE;
static bool firmware_hacks_available = TRUE, toppy_mode = FALSE, external_mode = FALSE, radio_mode = FALSE, radio_playback = FALSE, config_mode = FALSE;
static bool options_displayed = FALSE, key_config_displayed = FALSE, key_choice_displayed = FALSE, window_config_displayed = FALSE, options_changed;
static word region;
static dword state, sub_state, channel_clear_name_tick = 0, config_mode_clear_tick = 0, saved_config_clear_tick = 0, border_clear_tick = 0, toppy_mode_tick = 0, toppy_mode_delay;
static dword key_selection;
static int toppy_transparency;
static word current_pip_index, current_reverse_pip_index;

// bdb code
int main_x = 10, main_y = 576/4, main_w = SPLIT_W, main_h = SPLIT_H;
int sub_x = 720/2+5, sub_y = 576/4, sub_w = SPLIT_W, sub_h = SPLIT_H;
int x, y, w, h;

static int pre_radio_main_x, pre_radio_main_y, pre_radio_main_w, pre_radio_main_h;
static int pre_radio_sub_x, pre_radio_sub_y, pre_radio_sub_w, pre_radio_sub_h;
static int main_channel_number, channel_number, saved_radio_number = -1, saved_main_number, saved_pip_number = -1, tv_radio;

// bdb code
int sub_channel_number;

static int number_of_tv_channels, number_of_radio_channels, number_of_channels;
static int current_tv_channel_number = -1, previous_tv_channel_number = -1;
static int current_radio_channel_number = -1, previous_radio_channel_number = -1;
static char buffer[100];

static TYPE_TapEvent *main_channel_event, *sub_channel_event;
static TYPE_TapChInfo main_channel_info, sub_channel_info, channel_info, rec_channel_info_1, rec_channel_info_2;
static TYPE_PlayInfo play_info;
static TYPE_RecInfo rec_info_1, rec_info_2;

#define NUMBER_OF_SCALES 16
static word w_scales[NUMBER_OF_SCALES] = {SPLIT_W*2, SPLIT_W*2-(SPLIT_W/4), SPLIT_W+(SPLIT_W/2), SPLIT_W+(SPLIT_W/4), SPLIT_W+(SPLIT_W/8), SPLIT_W+(SPLIT_W/16), SPLIT_W, SPLIT_W-SPLIT_W/16, SPLIT_W-SPLIT_W/8, SPLIT_W/2+SPLIT_W/4, SPLIT_W/2+SPLIT_W/8, SPLIT_W/2, SPLIT_W/4+SPLIT_W/8, SPLIT_W/4, SPLIT_W/8+SPLIT_W/16, SPLIT_W/8};
static word h_scales[NUMBER_OF_SCALES] = {SPLIT_H*2, SPLIT_H*2-(SPLIT_H/4), SPLIT_H+(SPLIT_H/2), SPLIT_H+(SPLIT_H/4), SPLIT_H+(SPLIT_H/8), SPLIT_H+(SPLIT_H/16), SPLIT_H, SPLIT_H-SPLIT_H/16, SPLIT_H-SPLIT_H/8, SPLIT_H/2+SPLIT_H/4, SPLIT_H/2+SPLIT_H/8, SPLIT_H/2, SPLIT_H/4+SPLIT_H/8, SPLIT_H/4, SPLIT_H/8+SPLIT_H/16, SPLIT_H/8}; 
static int main_scale_index = 6, sub_scale_index = 6, scale_index;

//static word frame_colour				= RGB8888(150,150,150); // grey
//static word sub_info_colour			= RGB8888(181,18,18);	// red
//static word main_info_colour			= RGB8888(2,63,126);	// blue
//static word saved_settings_colour		= RGB8888(255,75,15);	// orange
//static word config_mode_colour		= RGB8888(32,162,32);	// green
//static word text_colour				= RGB8888(220,220,220); // white

static TYPE_Window options_window;

#ifdef WIN32
void * (*TAP_GetSystemProc)(int);
#endif
#ifndef WIN32
extern void * (*TAP_GetSystemProc)(int);
#endif*

int (*TAP_Channel_SwapMainSub)(void);
#define oTAP_Channel_SwapMainSub 0x15

// Return the Topfield model number
bool isTF5800PVRt()
{
	static word* sysID = (word*)0xa3fffffa;

#ifdef WIN32
	return TRUE;
#endif               

    switch (*sysID)
    {
		case 456:	return TRUE;
		case 458:	return TRUE;
		default:	return FALSE;
    }

    return FALSE;
}

// Return the Topfield model number
bool isTF5810PVRt()
{
	static word* sysID = (word*)0xa3fffffa;

#ifdef WIN32
	return TRUE;
#endif               

    switch (*sysID)
    {
		case 457:	return TRUE;
		default:	return FALSE;
    }

    return FALSE;
}

// Return the Topfield model number
bool isTF5000PVRt()
{
	return !isTF5800PVRt && !isTF5810PVRt();
}

bool isPipKey(dword param1)
{
	return ((param1 == RKEY_Red && (topfield_model == TF5000 || isTF5000PVRt()))
			|| (param1 == RKEY_Pip && isTF5800PVRt())
			|| (param1 == 0x10023 && isTF5810PVRt()));
}

bool useFirmwareHacks()
{
	return firmware_hacks_enabled && firmware_hacks_available;
}

void drawRadioIcon()
{
	if (display_icons && radio_mode) 
		TAP_Osd_PutGd(region, 50, 40, &_radio2Gd, TRUE);
}

void drawCurrentWindowIcon()
{
	if (display_icons == 2) {
		if (!radio_mode && main_x > 7 && main_y > 2 && (main_x + main_w + 10 + 4) < 720 && (main_y + main_h + 4) < 576)
			TAP_Osd_DrawRectangle(0, main_x-5, main_y, main_w+10, main_h, 2, frame_colour);
	}
	ef (display_icons == 1) {
		if (main_w > SPLIT_W/4 && !radio_mode) {
			if (main_h > 500)
				TAP_Osd_PutGd(region, 50, 50, &_soundGd, TRUE);
			else
				TAP_Osd_PutGd(region, (main_x+10 < 35) ? 35 : main_x+10, (main_y-45 < 20) ? main_y+main_h+15 : main_y-45, &_soundGd, TRUE);
		}
	}
}

void clearOsd()
{
	TAP_Osd_FillBox(region, 0, 0, 720, 576, COLOR_None);
	TAP_Osd_FillBox(0, 0, 0, 720, 576, COLOR_None);
	if (debug) TAP_Print("Clear Osd\r\n");
}

void setDefaultSettings()
{
	word i, j;

	main_x = startup_main_x, main_y = startup_main_y, main_w = startup_main_w, main_h = startup_main_h;
	sub_x = startup_sub_x, sub_y = startup_sub_y, sub_w = startup_sub_w, sub_h = startup_sub_h;

	main_scale_index = -1; sub_scale_index = -1;
	for (i = 0; i < NUMBER_OF_SCALES; i++)	{
		if (startup_main_w == w_scales[i]) main_scale_index = i;
		if (startup_sub_w == w_scales[i])  sub_scale_index = i;
	}

	for (i = 0; i < CURRENT_CONFIG_KEY_COUNT; i++) {
		for (j = 0; j < KEY_COUNT; j++) {
			if (config_keys[i].key_value == keys[j].key_value)
				current_config_keys[i].key_index = j;
		}
	}

	for (i = 0; i < CURRENT_NORMAL_KEY_COUNT; i++) {
		for (j = 0; j < KEY_COUNT; j++) {
			if (normal_keys[i].key_value == keys[j].key_value)
				current_normal_keys[i].key_index = j;
		}
	}

	for (i = 0; i < KEY_COUNT; i++) {
		if (pip_key == keys[i].key_value)
			current_pip_index = i;
		ef (reverse_pip_key == keys[i].key_value)
			current_reverse_pip_index = i;
	}

	if (debug) TAP_Print("Activation key is %s\r\n", keys[current_pip_index].key_name);
}

void saveCurrentSettings()
{
	startup_main_x = main_x, startup_main_y = main_y, startup_main_w = main_w, startup_main_h = main_h;
	startup_sub_x = sub_x, startup_sub_y = sub_y, startup_sub_w = sub_w, startup_sub_h = sub_h;

	SaveConfigurationToFile();
}

void scaleMain()
{
#ifndef WIN32
	TAP_Channel_Scale(CHANNEL_Main, main_x, main_y, main_w, main_h, TRUE);
	if (debug) TAP_Print("Scale main %d, %d, %d, %d\r\n", main_x, main_y, main_w, main_h);
#endif
}

void scaleSub()
{
#ifndef WIN32
	TAP_Channel_Scale(CHANNEL_Sub, sub_x, sub_y, sub_w, sub_h, TRUE);
	if (debug) TAP_Print("Scale sub %d, %d, %d, %d\r\n", sub_x, sub_y, sub_w, sub_h);
#endif
}

void scaleMainAndSub()
{
	if (active) {
		scaleMain();
		scaleSub();
	}
}

void drawChannelName(bool main)
{
	dword string_length, indent;
	word offset_x, channel_offset_y, event_offset_y;

	if (main) {
		// get play/channel info
		Safe_TAP_Hdd_GetPlayInfo(&play_info);
		TAP_Channel_GetCurrent(&tv_radio, &main_channel_number);
		TAP_Channel_GetInfo(tv_radio, main_channel_number, &main_channel_info);
		main_channel_event = TAP_GetCurrentEvent(tv_radio, main_channel_number);

		// populate channel name/event name
		if (play_info.playMode == PLAYMODE_Playing || play_info.playMode == PLAYMODE_RecPlaying)
			strcpy(buffer, play_info.evtInfo.eventName);
		else
			strcpy(buffer, main_channel_info.chName);

		// append TS if timeshifting
		if (play_info.playMode == PLAYMODE_TempPlaying)
			strcat(buffer, " - TS");

		// calculate the x,y offsets
		if (main_w >= SPLIT_W*2) {
			offset_x = ((720-10) / 2) - (170 / 2);
			
			if (main_channel_event 
					&& play_info.playMode != PLAYMODE_Playing 
					&& play_info.playMode != PLAYMODE_RecPlaying 
					&& play_info.playMode != PLAYMODE_TempPlaying) {
				channel_offset_y = 576-63-40;
				event_offset_y = channel_offset_y+40;
			}
			else
				channel_offset_y = 576-63;
		}
		ef (radio_mode) {
			offset_x = 720-230;
			channel_offset_y = 57;
			event_offset_y = channel_offset_y+40;
		}
		else {
			int indent = main_w/2-(170/2);
			
			offset_x = ((main_x+indent+170+2+20 > 720-30) ? (720-30) - (170+2+20) : main_x + indent < 50 ? 50 : main_x + indent);

			channel_offset_y = ((main_y-13) < 30) ? 30 : main_y-13;
			event_offset_y = ((channel_offset_y+main_h+40) > 576-20) ? (576-20) - 40 : channel_offset_y+main_h;
			if (event_offset_y-channel_offset_y < 40) channel_offset_y = event_offset_y - 40;
		}

		event_offset_y += 4;
		channel_offset_y += 4;

		// get the length of string for centralizing
		string_length = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
		if (string_length > (170-5)) string_length = 170-5;
		indent = ((170 - string_length) / 2);

		// draw the Osd
		if (play_info.playMode == PLAYMODE_Playing || play_info.playMode == PLAYMODE_RecPlaying || play_info.playMode == PLAYMODE_TempPlaying) {
			if (masterpiece_mode) {
				TAP_Osd_FillBox(0, offset_x, event_offset_y-5-7, 170, 40, COLOR_None);
				TAP_Osd_DrawRectangle(0, offset_x, event_offset_y-7, 170, 30, 2, 0x50);
				TAP_Osd_FillBox(0, offset_x+2, event_offset_y-7+2, 170-4, 30-4, 0x15);
				TAP_Osd_PutS(0, offset_x+indent, event_offset_y-3, offset_x+170-5, buffer, 0x37, COLOR_None, FALSE, FNT_Size_1622, FALSE, FALSE);
			}
			else {
				TAP_Osd_FillBox(region, offset_x-8, event_offset_y-5, 170+16, 40, COLOR_None);
				drawRoundedRectangle(offset_x, event_offset_y, 170, 15, 3, frame_colour, main_info_colour);
				TAP_Osd_PutS(region, offset_x+indent, event_offset_y-3, offset_x+170-5, buffer, text_colour, COLOR_None, FALSE, FNT_Size_1622, FALSE, FALSE);
			}
		
		}
		else {
			if (masterpiece_mode) {
				TAP_Osd_FillBox(0, offset_x, channel_offset_y-5-7, 170, 40, COLOR_None);
				TAP_Osd_DrawRectangle(0, offset_x, channel_offset_y-7, 170, 30, 2, 0x50);
				TAP_Osd_FillBox(0, offset_x+2, channel_offset_y-7+2, 170-4, 30-4, 0x15);
				TAP_Osd_PutS(0, offset_x+indent, channel_offset_y-3, offset_x+170-5, buffer, 0x37, COLOR_None, FALSE, FNT_Size_1622, FALSE, FALSE);
			}
			else {
				TAP_Osd_FillBox(region, offset_x-8, channel_offset_y-5, 170+16, 40, COLOR_None);
				drawRoundedRectangle(offset_x, channel_offset_y, 170, 15, 3, frame_colour, main_info_colour);
				TAP_Osd_PutS(region, offset_x+indent, channel_offset_y-3, offset_x+170-5, buffer, text_colour, COLOR_None, FALSE, FNT_Size_1622, FALSE, FALSE);
			}
		
		}

		// if live TV, also display the event name below
		if (main_channel_event 
				&& play_info.playMode != PLAYMODE_Playing 
				&& play_info.playMode != PLAYMODE_RecPlaying 
				&& play_info.playMode != PLAYMODE_TempPlaying) {

			// get the length of string for centralizing
			string_length = TAP_Osd_GetW(main_channel_event->eventName, 0, FNT_Size_1622);
			if (string_length > (170-5)) string_length = 170-5;
			indent = ((170 - string_length) / 2);

			// draw the Osd
			if (masterpiece_mode) {
				TAP_Osd_FillBox(0, offset_x, event_offset_y-5-7, 170, 40, COLOR_None);
				TAP_Osd_DrawRectangle(0, offset_x, event_offset_y-7, 170, 30, 2, 0x50);
				TAP_Osd_FillBox(0, offset_x+2, event_offset_y-7+2, 170-4, 30-4, 0x15);
				TAP_Osd_PutS(0, offset_x+indent, event_offset_y-3, offset_x+170-5, main_channel_event->eventName, 0x37, COLOR_None, FALSE, FNT_Size_1622, FALSE, FALSE);
			}
			else {
				TAP_Osd_FillBox(region, offset_x-8, event_offset_y-5, 170+16, 40, COLOR_None);
				drawRoundedRectangle(offset_x, event_offset_y, 170, 15, 3, frame_colour, main_info_colour);
				TAP_Osd_PutS(region, offset_x+indent, event_offset_y-3, offset_x+170-5, main_channel_event->eventName, text_colour, COLOR_None, FALSE, FNT_Size_1622, FALSE, FALSE);
			}
		}
	}
	else {
		// get channel info
		TAP_Channel_GetCurrentSubTV(&sub_channel_number);
		TAP_Channel_GetInfo(SVC_TYPE_Tv, sub_channel_number, &sub_channel_info);
		sub_channel_event = TAP_GetCurrentEvent(SVC_TYPE_Tv, sub_channel_number);

		// get the length of string for centralizing
		string_length = TAP_Osd_GetW(sub_channel_info.chName, 0, FNT_Size_1622);
		if (string_length > (170-5)) string_length = 170-5;
		indent = ((170 - string_length) / 2);

		// calculate the x,y offsets
		if (sub_w >= SPLIT_W*2 || radio_mode) {
			offset_x = ((720-10) / 2) - (170 / 2);

			if (sub_channel_event) {
				channel_offset_y = 576-63-40;
				event_offset_y = channel_offset_y+40;
			}
			else
				channel_offset_y = 576-63;
		}		
		else {
			int indent = sub_w/2-(170/2);

			TAP_Print("sub_x %d, sub_w %d, indent %d\r\n", sub_x, sub_w, indent);
			offset_x = ((sub_x+indent+170+2+20 > 720-30) ? (720-30) - (170+2+20) : sub_x + indent < 50 ? 50 : sub_x + indent);

			channel_offset_y = ((sub_y-13) < 30) ? 30 : sub_y-13;
			event_offset_y = ((channel_offset_y+sub_h+40) > 576-20) ? (576-20) - 40 : channel_offset_y+sub_h;
			if (event_offset_y-channel_offset_y < 40) channel_offset_y = event_offset_y - 40;
		}

		event_offset_y += 4;
		channel_offset_y += 4;
		TAP_Print("offset_x %d, channel_offset_y %d, event_offset_y %d\r\n", offset_x, channel_offset_y, event_offset_y);

		// draw the Osd
		if (masterpiece_mode) {
			TAP_Osd_FillBox(0, offset_x, channel_offset_y-5-7, 170, 40, COLOR_None);
			TAP_Osd_DrawRectangle(0, offset_x, channel_offset_y-7, 170, 30, 2, 0x50);
			TAP_Osd_FillBox(0, offset_x+2, channel_offset_y-7+2, 170-4, 30-4, 0xc3);
			TAP_Osd_PutS(0, offset_x+indent, channel_offset_y-3, offset_x+170-5, sub_channel_info.chName, 0x37, COLOR_None, FALSE, FNT_Size_1622, FALSE, FALSE);
		}
		else {
			TAP_Osd_FillBox(region, offset_x-8, channel_offset_y-5, 170+16, 40, COLOR_None);
			TAP_Print("x %d, y %d\r\n", offset_x, channel_offset_y);
			drawRoundedRectangle(offset_x, channel_offset_y, 170, 15, 3, frame_colour, sub_info_colour);
			TAP_Osd_PutS(region, offset_x+indent, channel_offset_y-3, offset_x+170-5, sub_channel_info.chName, text_colour, COLOR_None, FALSE, FNT_Size_1622, FALSE, FALSE);
		}

		// also display the event name below
		if (sub_channel_event) {

			// get the length of string for centralizing
			string_length = TAP_Osd_GetW(sub_channel_event->eventName, 0, FNT_Size_1622);
			if (string_length > (170-5)) string_length = 170-5;
			indent = ((170 - string_length) / 2);

			// draw the Osd
			if (masterpiece_mode) {
				TAP_Osd_FillBox(0, offset_x, event_offset_y-5-7, 170, 40, COLOR_None);
				TAP_Osd_DrawRectangle(0, offset_x, event_offset_y-7, 170, 30, 2, 0x50);
				TAP_Osd_FillBox(0, offset_x+2, event_offset_y-7+2, 170-4, 30-4, 0xc3);
				TAP_Osd_PutS(0, offset_x+indent, event_offset_y-3, offset_x+170-5, sub_channel_event->eventName, 0x37, COLOR_None, FALSE, FNT_Size_1622, FALSE, FALSE);
			}
			else {
				TAP_Osd_FillBox(region, offset_x-8, event_offset_y-5, 170+16, 40, COLOR_None);
				TAP_Print("x %d, y %d\r\n", offset_x, channel_offset_y);
				drawRoundedRectangle(offset_x, event_offset_y, 170, 15, 3, frame_colour, sub_info_colour);
				TAP_Osd_PutS(region, offset_x+indent, event_offset_y-3, offset_x+170-5, sub_channel_event->eventName, text_colour, COLOR_None, FALSE, FNT_Size_1622, FALSE, FALSE);
			}
		}
	}

	channel_clear_name_tick = TAP_GetTick();
}

void drawTime()
{
	dword string_length, indent;
	word mjd;
	byte hour, min, sec;

	TAP_GetTime(&mjd, &hour, &min, &sec);
	TAP_SPrint(buffer, "%02d:%02d", hour, min);

	string_length = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
	indent = ((80 - string_length) / 2);

	if (masterpiece_mode) {
		TAP_Osd_FillBox(0, ((720-30)/2)-40-8, 60-5-7, 80+16, 30+8, COLOR_None);
		TAP_Osd_DrawRectangle(0, ((720-30)/2)-40, 60-5-7+4, 80, 30, 2, 0x50);
		TAP_Osd_FillBox(0, ((720-30)/2)-40+2, 60-5-7+4+2, 80-4, 30-4, 0x78);
		TAP_Osd_PutS(0, ((720-30)/2)-40+indent, 60-10+6, ((720-30)/2)-40+80-5, buffer, 0x37, COLOR_None, FALSE, FNT_Size_1622, FALSE, FALSE);
	}
	else {
		TAP_Osd_FillBox(region, ((720-30)/2)-40-8, 60-5+4, 80+16, 30, COLOR_None);
		drawRoundedRectangle(((720-30)/2)-40, 60-5+4, 80, 15, 3, frame_colour, config_mode_colour);
		TAP_Osd_PutS(region, ((720-30)/2)-40+indent, 60-10+6, ((720-30)/2)-40+80-5, buffer, text_colour, COLOR_None, FALSE, FNT_Size_1622, FALSE, FALSE);
	}
}

void drawConfigMode()
{
	dword string_length, indent;

	if (config_mode)
		strcpy(buffer, "Config mode ON");
	else
		strcpy(buffer, "Config mode OFF");
	string_length = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
	indent = ((180 - string_length) / 2);

	if (masterpiece_mode) {
		TAP_Osd_FillBox(0, ((720-30)/2)-90-8, 60-5-7, 180+16, 30+8, COLOR_None);
		TAP_Osd_DrawRectangle(0, ((720-30)/2)-90, 60-5-7+4, 180, 30, 2, 0x50);
		TAP_Osd_FillBox(0, ((720-30)/2)-90+2, 60-5-7+4+2, 180-4, 30-4, 0x78);
		TAP_Osd_PutS(0, ((720-30)/2)-90+indent, 60-10+6, ((720-30)/2)-40+180-5, buffer, 0x37, COLOR_None, FALSE, FNT_Size_1622, FALSE, FALSE);
	}
	else {
		TAP_Osd_FillBox(region, ((720-30)/2)-90-8, 60-5+4-8, 180+16, 30+16, COLOR_None);
		drawRoundedRectangle(((720-30)/2)-90, 60-5+4, 180, 15, 3, frame_colour, config_mode_colour);
		TAP_Osd_PutS(region, ((720-30)/2)-90+indent, 60-10+6, ((720-30)/2)-40+180-5, buffer, text_colour, COLOR_None, FALSE, FNT_Size_1622, FALSE, FALSE);
	}

	config_mode_clear_tick = TAP_GetTick();
}

void drawSavedConfig()
{
	dword string_length, indent;

	strcpy(buffer, "Saved settings");
	string_length = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
	indent = ((180 - string_length) / 2);

	clearOsd();

	if (masterpiece_mode) {
		TAP_Osd_FillBox(0, ((720-30)/2)-90-8, 250-5-7, 180+16, 30+8, COLOR_None);
		TAP_Osd_DrawRectangle(0, ((720-30)/2)-90, 250-5-7+4, 180, 30, 2, 0x50);
		TAP_Osd_FillBox(0, ((720-30)/2)-90+2, 250-5-7+4+2, 180-4, 30-4, 0xe9);
		TAP_Osd_PutS(0, ((720-30)/2)-90+indent, 250-10+6, ((720-30)/2)-40+180-5, buffer, 0x37, COLOR_None, FALSE, FNT_Size_1622, FALSE, FALSE);
	}
	else {
		TAP_Osd_FillBox(region, ((720-30)/2)-90-8, 250-5+4-8, 180+16, 30+16, COLOR_None);
		drawRoundedRectangle(((720-30)/2)-90, 250-5+4, 180, 15, 3, frame_colour, saved_settings_colour);
		TAP_Osd_PutS(region, ((720-30)/2)-90+indent, 250-10+6, ((720-30)/2)-40+180-5, buffer, text_colour, COLOR_None, FALSE, FNT_Size_1622, FALSE, FALSE);
	}

	saved_config_clear_tick = TAP_GetTick();
}

void refreshDisplay()
{
	clearOsd();
	if (config_mode)
		drawConfigMode();
	drawCurrentWindowIcon();

	border_clear_tick = TAP_GetTick();
}

char *getCurrentKeyName(int current_key_index, bool edit_config_keys)
{
	if (edit_config_keys)
		return keys[current_config_keys[current_key_index].key_index].key_name;
	else
		return keys[current_normal_keys[current_key_index].key_index].key_name;
}

int getCurrentKeyValue(int current_key_index, bool edit_config_keys)
{
	if (edit_config_keys)
		return keys[current_config_keys[current_key_index].key_index].key_value;
	else
		return keys[current_normal_keys[current_key_index].key_index].key_value;
}

int getKeyIndex(int current_key_index, bool edit_config_keys)
{
	if (edit_config_keys)
		return current_config_keys[current_key_index].key_index;
	else
		return current_normal_keys[current_key_index].key_index;
}

void getKeyText(char *buffer, int current_key_index)
{
	strcpy(buffer, getCurrentKeyName(current_key_index, edit_config_keys));
}

/*bool keyNotInUse(int current_key_index, int key_index)
{
	word i, j;

	for (i = 0; i < CURRENT_CONFIG_KEY_COUNT; i++) {
		for (j = 0; j < KEY_COUNT; j++) {
			if (i != current_key_index && getCurrentKeyValue(current_key_index) == keys[key_index].key_value)
				return FALSE;
		}
	}

	return TRUE;
}*/

void drawCurrentWindowSettings()
{
	dword offset_x = 425, offset_y = 195;

	TAP_SPrint(buffer, "%d", startup_main_x);
	TAP_Osd_PutStringAf1622(0, offset_x, offset_y+(22*0), -1, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);
	TAP_SPrint(buffer, "%d", startup_main_y);
	TAP_Osd_PutStringAf1622(0, offset_x, offset_y+(22*1), -1, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);
	TAP_SPrint(buffer, "%d", startup_main_w);
	TAP_Osd_PutStringAf1622(0, offset_x, offset_y+(22*2), -1, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);
	TAP_SPrint(buffer, "%d", startup_main_h);
	TAP_Osd_PutStringAf1622(0, offset_x, offset_y+(22*3), -1, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);
	TAP_SPrint(buffer, "%d", startup_sub_x);
	TAP_Osd_PutStringAf1622(0, offset_x, offset_y+(22*4), -1, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);
	TAP_SPrint(buffer, "%d", startup_sub_y);
	TAP_Osd_PutStringAf1622(0, offset_x, offset_y+(22*5), -1, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);
	TAP_SPrint(buffer, "%d", startup_sub_w);
	TAP_Osd_PutStringAf1622(0, offset_x, offset_y+(22*6), -1, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);
	TAP_SPrint(buffer, "%d", startup_sub_h);
	TAP_Osd_PutStringAf1622(0, offset_x, offset_y+(22*7), -1, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);
}

void drawWindowSettingsWindow()
{
	clearOsd();

	if (debug) TAP_Print("Create window settings window\r\n");
	TAP_Win_Delete(&options_window);
	TAP_Win_SetDefaultColor(&options_window);
	TAP_Win_Create(&options_window, 0, 175, 140, 320, 220, FALSE, FALSE);

	TAP_Win_SetTitle(&options_window, "Default window settings", 0, FNT_Size_1622);
	TAP_Win_AddItem(&options_window, "Main window X Offset");
	TAP_Win_AddItem(&options_window, "Main window Y Offset");
	TAP_Win_AddItem(&options_window, "Main window width");
	TAP_Win_AddItem(&options_window, "Main window height");
	TAP_Win_AddItem(&options_window, "Sub window X Offset");
	TAP_Win_AddItem(&options_window, "Sub window Y Offset");
	TAP_Win_AddItem(&options_window, "Sub window width");
	TAP_Win_AddItem(&options_window, "Sub window height");
	TAP_Win_SetSelection(&options_window, 0);

	drawCurrentWindowSettings();
}

void drawChooseKeyWindow()
{
	word i;

	clearOsd();

	if (debug) TAP_Print("Create key choice window\r\n");
	TAP_Win_Delete(&options_window);
	TAP_Win_SetDefaultColor(&options_window);
	TAP_Win_Create(&options_window, 0, 205, 50, 230, 400, FALSE, TRUE);

	if (debug) TAP_Print("Choose a key\r\n");
	if (edit_config_keys)
		strcpy(buffer, current_config_keys[key_selection].key_description);
	else
		strcpy(buffer, current_normal_keys[key_selection].key_description);
	TAP_Win_SetTitle(&options_window, buffer, 0, FNT_Size_1622);
	//TAP_Win_SetFont(&options_window, 0, FNT_Size_1419);
	for (i = 0; i < KEY_COUNT; i++)
		TAP_Win_AddItem(&options_window, keys[i].key_name);

	TAP_Win_SetSelection(&options_window, getKeyIndex(key_selection, edit_config_keys));
}

void drawCurrentKeyValues()
{
	dword i, w, offset_y = 85;

	if (edit_config_keys) {
		for (i = 0; i < CURRENT_CONFIG_KEY_COUNT; i++) {
			getKeyText(buffer, i);
			w = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
			TAP_Osd_PutStringAf1622(0, 410, offset_y, 410+w, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);
			offset_y += 22;
		}
	}
	else {
		for (i = 0; i < CURRENT_NORMAL_KEY_COUNT; i++) {
			getKeyText(buffer, i);
			w = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
			TAP_Osd_PutStringAf1622(0, 410, offset_y, 410+w, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);
			offset_y += 22;
		}
	}
}

void drawConfigureKeysWindow(int key_index)
{
	word i;

	clearOsd();

	if (debug) TAP_Print("Create window\r\n");
	TAP_Win_Delete(&options_window);
	TAP_Win_SetDefaultColor(&options_window);
	TAP_Win_Create(&options_window, 0, 125, 30, 420, 420, FALSE, FALSE);

	if (edit_config_keys) {
		if (debug) TAP_Print("Config mode keys\r\n");
		TAP_Win_SetTitle(&options_window, "Config mode keys (Press OK to edit)", 0, FNT_Size_1622);
		for (i = 0; i < CURRENT_CONFIG_KEY_COUNT; i++)
			TAP_Win_AddItem(&options_window, current_config_keys[i].key_description);
	}
	else {
		if (debug) TAP_Print("Normal mode keys\r\n");
		TAP_Win_SetTitle(&options_window, "Normal mode keys (Press OK to edit)", 0, FNT_Size_1622);
		for (i = 0; i < CURRENT_NORMAL_KEY_COUNT; i++)
			TAP_Win_AddItem(&options_window, current_normal_keys[i].key_description);
	}

	TAP_Win_SetSelection(&options_window, key_index);

	drawCurrentKeyValues();
}

void setKeyText(dword key, char *buffer)
{
	word i;

	for (i = 0; i < KEY_COUNT; i++) {
		if (key == keys[i].key_value)
			strcpy(buffer, keys[i].key_name);
	}
}

void drawOptionValues()
{
	dword w, offset_y = 115;

	TAP_SPrint(buffer, "%s", "Press OK");
	w = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
	TAP_Osd_PutStringAf1622(0, 400, offset_y+(22*0), 400+w, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);

	TAP_SPrint(buffer, "%s", "Press OK");
	w = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
	TAP_Osd_PutStringAf1622(0, 400, offset_y+(22*1), 400+w, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);

	TAP_SPrint(buffer, "%s", "Press OK");
	w = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
	TAP_Osd_PutStringAf1622(0, 400, offset_y+(22*2), 400+w, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);

	setKeyText(pip_key, buffer);
	w = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
	TAP_Osd_PutStringAf1622(0, 400, offset_y+(22*3), 400+w, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);

	setKeyText(reverse_pip_key, buffer);
	w = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
	TAP_Osd_PutStringAf1622(0, 400, offset_y+(22*4), 400+w, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);

	TAP_SPrint(buffer, "%s", recall_snap ? "Saved settings" : "Split screen");
	w = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
	TAP_Osd_PutStringAf1622(0, 400, offset_y+(22*5), 400+w, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);

	TAP_SPrint(buffer, "%s", topfield_model == TF5800 ? "TF5800" : (topfield_model == TF5000 ? "TF5000" : "TF5810"));
	w = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
	TAP_Osd_PutStringAf1622(0, 400, offset_y+(22*6), 400+w, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);

	TAP_SPrint(buffer, "%s", masterpiece_mode ? "On" : "Off");
	w = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
	TAP_Osd_PutStringAf1622(0, 400, offset_y+(22*7), 400+w, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);

	TAP_SPrint(buffer, "%s", (display_icons == 0) ? "Off" : (display_icons == 1) ? "Icon" : "Border");
	w = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
	TAP_Osd_PutStringAf1622(0, 400, offset_y+(22*8), 400+w, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);

	if (info_delay == 0)
		strcpy(buffer, "Ever");
	else 
		TAP_SPrint(buffer, "%d %s", info_delay, (info_delay == 1) ? "second" : "seconds");
	w = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
	TAP_Osd_PutStringAf1622(0, 400, offset_y+(22*9), 400+w, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);

	TAP_SPrint(buffer, "%s", firmware_hacks_enabled ? "On" : "Off");
	w = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
	TAP_Osd_PutStringAf1622(0, 400, offset_y+(22*10), 400+w, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);

	memset(buffer, '\0', 100);
	if (current_fav == -1)
		strcpy(buffer, "None");
	else
		strncat(buffer, fav_file_name, strlen(fav_file_name)-4);
	w = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
	TAP_Osd_PutStringAf1622(0, 400, offset_y+(22*11), 400+w, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);

	memset(buffer, '\0', 100);
	strncat(buffer, skin_file_name, strlen(skin_file_name)-4);
	w = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
	TAP_Osd_PutStringAf1622(0, 400, offset_y+(22*12), 400+w, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);

	TAP_SPrint(buffer, "%s", "Press OK");
	w = TAP_Osd_GetW(buffer, 0, FNT_Size_1622);
	TAP_Osd_PutStringAf1622(0, 400, offset_y+(22*13), 400+w, buffer, masterpiece_mode ? 0x15 : RGB8888(0,0,204), COLOR_None);
}

void drawOptionsWindow(dword selection)
{
	clearOsd();

	if (debug) TAP_Print("Create window\r\n");
	TAP_Win_Delete(&options_window);
	TAP_Win_SetDefaultColor(&options_window);
	TAP_Win_Create(&options_window, 0, 150, 60, 370, 295, FALSE, FALSE);

	if (debug) TAP_Print("Super PiP Settings\r\n");
	TAP_Win_SetTitle(&options_window, "Super PiP Settings", 0, FNT_Size_1622);
	TAP_Win_AddItem(&options_window, "Define config keys");
	TAP_Win_AddItem(&options_window, "Define normal keys");
	TAP_Win_AddItem(&options_window, "Default window settings");
	TAP_Win_AddItem(&options_window, "PiP key");
	TAP_Win_AddItem(&options_window, "Reverse PiP key");
	TAP_SPrint(buffer, "'%s' snaps to", getCurrentKeyName(1, TRUE));
	TAP_Win_AddItem(&options_window, buffer);
	TAP_Win_AddItem(&options_window, "Topfield model");
	TAP_Win_AddItem(&options_window, "Masterpiece mode");
	TAP_Win_AddItem(&options_window, "Active window indicator");
	TAP_Win_AddItem(&options_window, "Display channel info for");
	TAP_Win_AddItem(&options_window, "Firmware hacks");
	TAP_Win_AddItem(&options_window, "Choose favourites");
	TAP_Win_AddItem(&options_window, "Choose skin");
	TAP_Win_AddItem(&options_window, "Exit Super PiP");
	TAP_Win_SetSelection(&options_window, selection);

	drawOptionValues();
}

bool includedInFavourites(int channel_number)
{
	word i;

	if (number_of_favourites == 0)
		return TRUE;

	TAP_Print("Check channel %d in favs\r\n", channel_number);
	for (i = 0; i < number_of_favourites; i++) {
		if (favourite_channels[i] == channel_number)
			return TRUE;
	}

	return FALSE;
}

bool isChannelNotPlayable(int channel_number,
						  bool recording_on_1, bool recording_on_2, 
						  TYPE_TapChInfo channel_info, TYPE_TapChInfo sub_channel_info, 
						  TYPE_TapChInfo rec_channel_info_1, TYPE_TapChInfo rec_channel_info_2)
{
	if (!includedInFavourites(channel_info.logicalChNum))
		return TRUE;

	return ((recording_on_1 && (sub_channel_info.freq != rec_channel_info_1.freq) && channel_info.freq != sub_channel_info.freq && channel_info.freq != rec_channel_info_1.freq)
			|| (recording_on_2 && (sub_channel_info.freq != rec_channel_info_2.freq) && channel_info.freq != sub_channel_info.freq && channel_info.freq != rec_channel_info_2.freq)
			|| ((recording_on_1 && recording_on_2) && (rec_channel_info_1.freq != rec_channel_info_2.freq) && channel_info.freq != rec_channel_info_1.freq && channel_info.freq != rec_channel_info_2.freq));
}

void checkChannelIsPlayable(int *channel_number, bool recording_on_1, bool recording_on_2, TYPE_RecInfo rec_info_1, TYPE_RecInfo rec_info_2,
							TYPE_TapChInfo channel_info, TYPE_TapChInfo sub_channel_info, TYPE_TapChInfo rec_channel_info_1, TYPE_TapChInfo rec_channel_info_2, 
							bool search_up, int tv_radio)
{
	number_of_channels = (tv_radio == SVC_TYPE_Tv) ? number_of_tv_channels : number_of_radio_channels;
	if (debug) TAP_Print("Check for non playable channels, total %d\r\n", number_of_channels);

	// get recording channel details
	if (recording_on_1) TAP_Channel_GetInfo(rec_info_1.svcType, rec_info_1.svcNum, &rec_channel_info_1);
	if (recording_on_2) TAP_Channel_GetInfo(rec_info_2.svcType, rec_info_2.svcNum, &rec_channel_info_2);

	// get info for channel to play
	TAP_Channel_GetInfo(tv_radio, *channel_number, &channel_info);

	// while it is not playable, get the next channel
	while (isChannelNotPlayable(*channel_number, recording_on_1, recording_on_2, channel_info, sub_channel_info, rec_channel_info_1, rec_channel_info_2)) {
		if (debug) TAP_Print("Found non playable channel %d\r\n", *channel_number);
		if (search_up)
			*channel_number = (*channel_number == (number_of_channels - 1)) ? 0 : *channel_number + 1;
		else
			*channel_number = (*channel_number == 0) ? number_of_channels - 1 : *channel_number - 1;
		TAP_Channel_GetInfo(SVC_TYPE_Tv, *channel_number, &channel_info);
	}
}

void changeChannelUp(bool main)
{
	bool recording_on_1 = Safe_TAP_Hdd_GetRecInfo(0, &rec_info_1) && ((rec_info_1.recType == RECTYPE_Normal) || (rec_info_1.recType == RECTYPE_Copy));
	bool recording_on_2 = Safe_TAP_Hdd_GetRecInfo(1, &rec_info_2) && ((rec_info_2.recType == RECTYPE_Normal) || (rec_info_2.recType == RECTYPE_Copy));

	if (main) {
		// abort if video playing
		Safe_TAP_Hdd_GetPlayInfo(&play_info);
		if (play_info.playMode == PLAYMODE_Playing || play_info.playMode == PLAYMODE_RecPlaying || play_info.playMode == PLAYMODE_TempPlaying)
			return;

		// get the current channel details
		TAP_Channel_GetCurrent(&tv_radio, &main_channel_number);
		TAP_Channel_GetCurrentSubTV(&sub_channel_number);
		TAP_Channel_GetInfo(SVC_TYPE_Tv, sub_channel_number, &sub_channel_info);
		number_of_channels = (tv_radio == SVC_TYPE_Tv) ? number_of_tv_channels : number_of_radio_channels;
		channel_number = main_channel_number;
		channel_number = (channel_number == (number_of_channels - 1)) ? 0 : channel_number + 1;

		//if (recording_on_1 || recording_on_2) {
			checkChannelIsPlayable(&channel_number, recording_on_1, recording_on_2, rec_info_1, rec_info_2, 
								   channel_info, sub_channel_info, rec_channel_info_1, rec_channel_info_2, 
								   TRUE, tv_radio);
		//}

		TAP_Channel_Start(CHANNEL_Main, tv_radio, channel_number);
		scaleMain();
	}
	else {
		// get the current channel details
		TAP_Channel_GetCurrent(&tv_radio, &main_channel_number);
		TAP_Channel_GetCurrentSubTV(&sub_channel_number);
		TAP_Channel_GetInfo(tv_radio, main_channel_number, &main_channel_info);
		channel_number = sub_channel_number;
		channel_number = (channel_number == (number_of_tv_channels - 1)) ? 0 : channel_number + 1;

		//if (recording_on_1 || recording_on_2) {
			checkChannelIsPlayable(&channel_number, recording_on_1, recording_on_2, rec_info_1, rec_info_2, 
									channel_info, main_channel_info, rec_channel_info_1, rec_channel_info_2,
									TRUE, SVC_TYPE_Tv);
		//}

		TAP_Channel_Start(CHANNEL_Sub, SVC_TYPE_Tv, channel_number);
		scaleSub();
	}

	clearOsd();
	if (config_mode)
		drawConfigMode();
	drawCurrentWindowIcon();
	drawRadioIcon();
	drawChannelName(main);

	if (debug) TAP_Print("Changed channel up to %d\r\n", channel_number);
}

void changeChannelDown(bool main)
{
	bool recording_on_1 = Safe_TAP_Hdd_GetRecInfo(0, &rec_info_1) && ((rec_info_1.recType == RECTYPE_Normal) || (rec_info_1.recType == RECTYPE_Copy));
	bool recording_on_2 = Safe_TAP_Hdd_GetRecInfo(1, &rec_info_2) && ((rec_info_2.recType == RECTYPE_Normal) || (rec_info_2.recType == RECTYPE_Copy));

	if (main) {
		// abort if video playing
		Safe_TAP_Hdd_GetPlayInfo(&play_info);
		if (play_info.playMode == PLAYMODE_Playing || play_info.playMode == PLAYMODE_RecPlaying || play_info.playMode == PLAYMODE_TempPlaying)
			return;

		// get the current channel details
		TAP_Channel_GetCurrent(&tv_radio, &main_channel_number);
		TAP_Channel_GetCurrentSubTV(&sub_channel_number);
		TAP_Channel_GetInfo(SVC_TYPE_Tv, sub_channel_number, &sub_channel_info);
		number_of_channels = (tv_radio == SVC_TYPE_Tv) ? number_of_tv_channels : number_of_radio_channels;
		channel_number = main_channel_number;
		channel_number = (channel_number == 0) ? number_of_channels - 1 : channel_number - 1;

		//if (recording_on_1 || recording_on_2) {
			checkChannelIsPlayable(&channel_number, recording_on_1, recording_on_2, rec_info_1, rec_info_2, 
								   channel_info, sub_channel_info, rec_channel_info_1, rec_channel_info_2, 
								   FALSE, tv_radio);
		//}

		TAP_Channel_Start(CHANNEL_Main, tv_radio, channel_number);
		scaleMain();
	}
	else {
		// get the current channel details
		TAP_Channel_GetCurrent(&tv_radio, &main_channel_number);
		TAP_Channel_GetCurrentSubTV(&sub_channel_number);
		TAP_Channel_GetInfo(tv_radio, main_channel_number, &main_channel_info);
		channel_number = sub_channel_number;
		channel_number = (channel_number == 0) ? number_of_tv_channels - 1 : channel_number - 1;

		//if (recording_on_1 || recording_on_2) {
			checkChannelIsPlayable(&channel_number, recording_on_1, recording_on_2, rec_info_1, rec_info_2, 
								   channel_info, main_channel_info, rec_channel_info_1, rec_channel_info_2, 
								   FALSE, SVC_TYPE_Tv);
		//}

		TAP_Channel_Start(CHANNEL_Sub, SVC_TYPE_Tv, channel_number);
		scaleSub();
	}

	clearOsd();
	if (config_mode)
		drawConfigMode();
	drawCurrentWindowIcon();
	drawRadioIcon();
	drawChannelName(main);
	if (debug) TAP_Print("Changed channel down to %d\r\n", channel_number);
}

void shiftChannelUp(bool main)
{
	if (radio_mode)
		return;

	if (main) {
		if (main_w >= SPLIT_W*2)
			return;

		main_y = ((main_y - 20) < 0) ? 576-10 - main_h : main_y - 10;
		scaleMain();
	}
	else {
		if (sub_w >= SPLIT_W*2)
			return;

		sub_y = ((sub_y - 15) < 0) ? 576-10 - sub_h : sub_y - 10;
		scaleSub();
	}

	refreshDisplay();
	if (debug) TAP_Print("Moved channel up\r\n");
}

void shiftChannelDown(bool main)
{
	if (radio_mode)
		return;

	if (main) {
		if (main_w >= SPLIT_W*2)
			return;

		main_y = ((main_y + 20) > 576 - main_h) ? 10 : main_y + 10;
		scaleMain();
	}
	else {
		if (sub_w >= SPLIT_W*2)
			return;

		sub_y = ((sub_y + 15) > 576 - sub_h) ? 10 : sub_y + 10;
		scaleSub();
	}

	refreshDisplay();
	if (debug) TAP_Print("Moved channel down\r\n");
}

void shiftChannelLeft(bool main)
{
	if (radio_mode)
		return;

	if (main) {
		if (main_w >= SPLIT_W*2)
			return;

		main_x = ((main_x - 20) < 0) ? 720-10 - main_w : main_x - 10;
		scaleMain();
	}
	else {
		if (sub_w >= SPLIT_W*2)
			return;

		sub_x = ((sub_x - 15) < 0) ? 720-10 - sub_w : sub_x - 10;
		scaleSub();
	}

	refreshDisplay();
	if (debug) TAP_Print("Moved channel left\r\n");
}

void shiftChannelRight(bool main)
{
	if (radio_mode)
		return;

	if (main) {
		if (main_w >= SPLIT_W*2)
			return;

		main_x = ((main_x + 20) > 720 - main_w) ? 10 : main_x + 10;
		scaleMain();
	}
	else {
		if (sub_w >= SPLIT_W*2)
			return;

		sub_x = ((sub_x + 15) > 720 - sub_w) ? 10 : sub_x + 10;
		scaleSub();
	}

	refreshDisplay();
	if (debug) TAP_Print("Move channel right\r\n");
}

bool isPredefinedScale(bool main)
{
	word i;

	for (i = 0; i < NUMBER_OF_SCALES; i++) {
		if (main) { 
			if (main_w == w_scales[i] && main_h == h_scales[i])
				return TRUE;
		}
		else { 
			if (sub_w == w_scales[i] && sub_h == h_scales[i])
				return TRUE;
		}
	}

	return FALSE;
}

int getNearestScaleIndex(word width, word height, bool up)
{
	word i;

	// first check outside the predefined bounds
	if (up) {
		if (width > w_scales[0])
			return -1;
		ef (width < w_scales[NUMBER_OF_SCALES-1])
			return NUMBER_OF_SCALES-1;
	}
	else {
		if (width < w_scales[NUMBER_OF_SCALES-1])
			return -1;
		ef (width > w_scales[0])
			return 0;		
	}

	// must be within predefined bounds, so shuffle to nearest
	TAP_Print("Check %d\r\n", width);
	for (i = 0; i < NUMBER_OF_SCALES-1; i++) {
		if (up) { 
			if (width > w_scales[i+1] && width < w_scales[i])
				return i;
		}
		else { 
			TAP_Print("Check %d against %d & %d\r\n", width, w_scales[i], w_scales[i+1]);
			if (width > w_scales[i+1] && width < w_scales[i])
				return i+1;
		}
	}

	// should never get here, but just in case
	TAP_Print("Return default\r\n");
	return up ? 0 : NUMBER_OF_SCALES-1;
}

int getScaleIndex(bool main, bool up)
{
	int new_scale;

	if (main) {
		if (up) {
			if (isPredefinedScale(TRUE)) {
				if (main_scale_index == 0)
					return -1;

				new_scale = main_scale_index-1;
			}
			else {
				new_scale = getNearestScaleIndex(main_w, main_h, TRUE);
			}

			return new_scale;
		}
		else {
			if (isPredefinedScale(TRUE)) {
				if (main_scale_index >= NUMBER_OF_SCALES-1)
					return -1;

				new_scale = main_scale_index+1;
			}
			else {
				new_scale = getNearestScaleIndex(main_w, main_h, FALSE);
			}

			return new_scale;
		}
	}
	else {
		if (up) {
			if (isPredefinedScale(FALSE)) {
				if (sub_scale_index == 0)
					return -1;

				new_scale = sub_scale_index-1;
			}
			else {
				new_scale = getNearestScaleIndex(sub_w, sub_h, TRUE);
			}

			return new_scale;
		}
		else {
			if (isPredefinedScale(FALSE)) {
				if (sub_scale_index >= NUMBER_OF_SCALES-1)
					return -1;

				new_scale = sub_scale_index+1;
			}
			else {
				new_scale = getNearestScaleIndex(sub_w, sub_h, FALSE);
			}

			return new_scale;
		}
	}
}

void shrinkChannel(bool main)
{
	int width_adjust, height_adjust, new_scale;

	if (radio_mode)
		return;

	if (main) {
		new_scale = getScaleIndex(TRUE, FALSE);
		if (new_scale < 0)
			return;

		main_scale_index = new_scale;
		width_adjust = (w_scales[main_scale_index-1] - w_scales[main_scale_index]) / 2;
		height_adjust = (h_scales[main_scale_index-1] - h_scales[main_scale_index]) / 2;
		main_x += width_adjust;
		main_y += height_adjust;

		main_w = w_scales[main_scale_index];
		main_h = h_scales[main_scale_index];

		scaleMain();
	}
	else {
		new_scale = getScaleIndex(FALSE, FALSE);
		if (new_scale < 0)
			return;

		sub_scale_index = new_scale;
		width_adjust = (w_scales[sub_scale_index-1] - w_scales[sub_scale_index]) / 2;
		height_adjust = (h_scales[sub_scale_index-1] - h_scales[sub_scale_index]) / 2;
		sub_x += width_adjust;
		sub_y += height_adjust;

		sub_w = w_scales[sub_scale_index];
		sub_h = h_scales[sub_scale_index];

		scaleSub();
	}

	refreshDisplay();
	if (debug) TAP_Print("Shrunk channel\r\n");
}

void enlargeChannel(bool main)
{
	int new_start_x, new_start_y, width_adjust, new_scale;
	int new_end_x, new_end_y, height_adjust;

	if (radio_mode)
		return;

	if (main) {
		new_scale = getScaleIndex(TRUE, TRUE);
		if (new_scale < 0)
			return;

		// calculate new start and end points
		main_scale_index = new_scale;
		width_adjust = (w_scales[main_scale_index] - w_scales[main_scale_index+1]) / 2;
		height_adjust = (h_scales[main_scale_index] - h_scales[main_scale_index+1]) / 2;
		new_start_x = (main_x - width_adjust), new_start_y = (main_y - height_adjust);
		new_end_x = new_start_x + (main_w + width_adjust), new_end_y = new_start_y + (main_h + height_adjust);

		main_w = w_scales[main_scale_index];
		main_h = h_scales[main_scale_index];

		// if still within screen, then set to new start point, otherwise set to bounds
		if (new_start_x > 10 && new_end_x < 720-10) 
			main_x = new_start_x;
		else
			main_x = (new_start_x <= 10) ? 10 : 720-10 - main_w;

		// if still within screen, then set to new start point, otherwise set to bounds
		if (new_start_y > 10 && new_end_y < 576-10) 
			main_y = new_start_y;
		else
			main_y = (new_start_y <= 10) ? 10 : 576-10 - main_h;
		
		scaleMain();
	}
	else {
		new_scale = getScaleIndex(FALSE, TRUE);
		if (new_scale < 0)
			return;

		// calculate new start and end points
		sub_scale_index = new_scale;
		width_adjust = (w_scales[sub_scale_index] - w_scales[sub_scale_index+1]) / 2;
		height_adjust = (h_scales[sub_scale_index] - h_scales[sub_scale_index+1]) / 2;
		new_start_x = (sub_x - width_adjust), new_start_y = (sub_y - height_adjust);
		new_end_x = new_start_x + (sub_w + width_adjust), new_end_y = new_start_y + (sub_h + height_adjust);

		sub_w = w_scales[sub_scale_index];
		sub_h = h_scales[sub_scale_index];

		// if still within screen, then set to new start point, otherwise set to bounds
		if (new_start_x > 10 && new_end_x < 720-10) 
			sub_x = new_start_x;
		else
			sub_x = (new_start_x <= 10) ? 10 : 720-10 - sub_w;

		// if still within screen, then set to new start point, otherwise set to bounds
		if (new_start_y > 10 && new_end_y < 576-10) 
			sub_y = new_start_y;
		else
			sub_y = (new_start_y <= 10) ? 10 : 576-10 - sub_h;

		scaleSub();
	}

	refreshDisplay();
	if (debug) TAP_Print("Enlarged channel\r\n");
}

int getSubChannelStartNumber()
{
	if (previous_tv_channel_number >= 0 && (previous_tv_channel_number != main_channel_number) && (first_activation || (saved_pip_number == main_channel_number)))
		return previous_tv_channel_number;
	ef (saved_pip_number >= 0 && (saved_pip_number != main_channel_number))
		return saved_pip_number;
	ef (sub_channel_number >= 0 && (sub_channel_number != main_channel_number))
		return sub_channel_number;
	ef (main_channel_number == 0)
		return 1;

	return 0;
}

int getRadioStartNumber()
{
	if (saved_radio_number >= 0)
		return saved_radio_number;
	ef (current_radio_channel_number >= 0)
		return current_radio_channel_number;

	return 0;
}

void handleExitTap(bool close_pip)
{
#ifndef WIN32
	// bdb code
	if (useFirmwareHacks())
		close_exTras();
#endif

	TAP_Win_Delete(&options_window);
	if (close_pip) {
		clearOsd();
		TAP_Osd_Delete(region);

		TAP_Channel_Start(CHANNEL_Sub, 0, sub_channel_number);
		TAP_Channel_Stop(CHANNEL_Sub);
		TAP_Channel_Scale(CHANNEL_Main, 0, 0, 720, 576, TRUE);
		if (debug) TAP_Print("Show full main screen\r\n");

	}

	TAP_EnterNormal();
	TAP_Exit();
	if (debug) TAP_Print("Exit tap\r\n");
}

void handleClosePip(bool show_main)
{
#ifndef WIN32
	// bdb code
	if (useFirmwareHacks())
		close_exTras();
#endif
	clearOsd();
	TAP_Osd_Delete(region);

	TAP_Channel_GetCurrent(&tv_radio, &main_channel_number);
	TAP_Channel_GetCurrentSubTV(&sub_channel_number);

	if (radio_mode) {
		main_x = pre_radio_main_x, main_y = pre_radio_main_y, main_w = pre_radio_main_w, main_h = pre_radio_main_h;
		sub_x = pre_radio_sub_x, sub_y = pre_radio_sub_y, sub_w = pre_radio_sub_w, sub_h = pre_radio_sub_h;
		saved_radio_number = main_channel_number;
		radio_mode = FALSE;
	}
	else {
		saved_pip_number = reverse_pip ? main_channel_number : sub_channel_number;
	}

	if (show_main) {
		TAP_Channel_Start(CHANNEL_Sub, 0, sub_channel_number);
		TAP_Channel_Stop(CHANNEL_Sub);
		TAP_Channel_Scale(CHANNEL_Main, 0, 0, 720, 576, TRUE);
		if (debug) TAP_Print("Show full main screen\r\n");
	}
	else {
		TAP_Channel_Stop(CHANNEL_Sub);
		TAP_Channel_Start(CHANNEL_Main, 0, sub_channel_number);
		TAP_Channel_Scale(CHANNEL_Main, 0, 0, 720, 576, TRUE);
		if (debug) TAP_Print("Shown full sub screen\r\n");
	}

	TAP_EnterNormal();
	active = FALSE;
	reverse_pip = FALSE;
	config_mode = FALSE;
}

dword handlePipKey(dword param1, bool reverse)
{
	// get the current channels
	TAP_Channel_GetCurrent(&tv_radio, &main_channel_number);
	TAP_Channel_GetCurrentSubTV(&sub_channel_number);
					
	if (!active) {
		// ignore if already PiP
		if (TAP_Channel_IsStarted(0)) 
			return param1;

		// check state
		TAP_GetState(&state, &sub_state);
		if (debug) TAP_Print("State %d, Substate %d\r\n", state, sub_state);
		if (state != STATE_Normal || sub_state != SUBSTATE_Normal)
			return param1;

		// does not make sense to start reverse pip while in radio or playback
		if (reverse) {
			
			Safe_TAP_Hdd_GetPlayInfo(&play_info);
			if (play_info.playMode == PLAYMODE_Playing || play_info.playMode == PLAYMODE_RecPlaying)
				return param1;

			if (tv_radio == SVC_TYPE_Radio)
				return param1;

			reverse_pip = TRUE;
		}

		if (isPipKey(param1)) {
			// once we have the pip key, pass it on and send out the super pip activation 
			active = TRUE;
			pip_generated = FALSE;
			TAP_GenerateEvent(EVT_KEY, Super_PiP, 0);
			if (debug) TAP_Print("Activate Super PiP\r\n");
			return param1;
		}
		else {
			// if the activation key is not pip, then generate it
			pip_generated = TRUE;
			TAP_GenerateEvent(EVT_KEY, (topfield_model == TF5000 || isTF5000PVRt()) ? RKEY_Red : (topfield_model == TF5800 ? RKEY_Pip : 0x10023), 0);
			if (debug) TAP_Print("Simulate Toppy PiP\r\n");
		}
	}
	else {
		handleClosePip(reverse_pip ? FALSE : TRUE);
	}

	return 0;
}

void handleInfoKey()
{
	clearOsd();
	if (config_mode)
		drawConfigMode();
	drawCurrentWindowIcon();
	drawRadioIcon();
	drawChannelName(TRUE);
	drawChannelName(FALSE);
	//drawTime();
}

bool isRadioRecording(TYPE_PlayInfo play_info)
{
	if ((play_info.playMode == PLAYMODE_Playing || play_info.playMode == PLAYMODE_RecPlaying)
			&& (play_info.svcType == SVC_TYPE_Radio))
				//|| (isTF5800PVRt() 
				//	&& TAP_Channel_GetInfo(play_info.svcType, play_info.svcNum, &channel_info) == 0
				//	&& channel_info.logicalChNum >= 700 && channel_info.logicalChNum < 800)))
		return TRUE;

	return FALSE;
}

void handleSuperPipKey()
{
	if (debug) TAP_Print("Super pip received\r\n");

#ifndef WIN32
	// if pip not started, then abandon and reset (should have been started by Toppy)
	if (!TAP_Channel_IsStarted(CHANNEL_Sub)) {
		if (debug) TAP_Print("PiP not started, so abort Super PiP mode\r\n");
		TAP_EnterNormal();
		active = FALSE;
		reverse_pip = FALSE;
		return;
	}

	// bdb code
	if (debug) TAP_Print("Enable hacks\r\n");
	if (useFirmwareHacks())
		open_exTras();
#endif
	TAP_ExitNormal();

	// test for a currently playing radio recording
	radio_playback = FALSE;
	Safe_TAP_Hdd_GetPlayInfo(&play_info);
	if (isRadioRecording(play_info)) {
		if (debug) TAP_Print("Radio playback in progress\r\n");
		radio_playback = TRUE;
	}

	// get the current channels
	TAP_Channel_GetCurrent(&tv_radio, &main_channel_number);
	TAP_Channel_GetCurrentSubTV(&sub_channel_number);
	
	region = TAP_Osd_Create(0, 0, 720, 576, 0, FALSE);

	if (!radio_playback && tv_radio == SVC_TYPE_Tv) {
		if (reverse_pip) {
			if (debug) TAP_Print("Start main %d\r\n", getSubChannelStartNumber());
			TAP_Channel_Start(CHANNEL_Main, SVC_TYPE_Tv, getSubChannelStartNumber());

			if (debug) TAP_Print("Start sub %d\r\n", main_channel_number);
			TAP_Channel_Start(CHANNEL_Sub, SVC_TYPE_Tv, main_channel_number);
		}
		else {
			if (debug) TAP_Print("Start sub %d\r\n", getSubChannelStartNumber());
			TAP_Channel_Start(CHANNEL_Sub, SVC_TYPE_Tv, getSubChannelStartNumber());
		}
	}
	else {
		pre_radio_main_x = main_x, pre_radio_main_y = main_y, pre_radio_main_w = main_w, pre_radio_main_h = main_h;
		pre_radio_sub_x = sub_x, pre_radio_sub_y = sub_y, pre_radio_sub_w = sub_w, pre_radio_sub_h = sub_h;
		main_x = 720-250, main_y = 100, main_w = w_scales[NUMBER_OF_SCALES-1], main_h = h_scales[NUMBER_OF_SCALES-1];
		sub_x = 10, sub_y = 10, sub_w = w_scales[0], sub_h = h_scales[0];
		TAP_Channel_Start(CHANNEL_Sub, SVC_TYPE_Tv, current_tv_channel_number >= 0 ? current_tv_channel_number : 0);

		saved_main_number = main_channel_number = current_tv_channel_number >= 0 ? current_tv_channel_number : 0;
		saved_pip_number = getSubChannelStartNumber();
		radio_mode = TRUE;
	}
	
	scaleMainAndSub();
	scaleMainAndSub(); 

	if (debug) TAP_Print("Draw channel info\r\n");
	handleInfoKey();

	first_activation = FALSE;
	if (debug) TAP_Print("Activated Super PiP\r\n");

}

void handlePipSwapKey()
{
	Safe_TAP_Hdd_GetPlayInfo(&play_info);
	if (play_info.playMode == PLAYMODE_Playing || play_info.playMode == PLAYMODE_RecPlaying || play_info.playMode == PLAYMODE_TempPlaying || radio_mode)
		return;
	
	// can use this to swap (but gives a brief full screen effect which is undesirable)
	//TAP_Channel_SwapMainSub = TAP_GetSystemProc(oTAP_Channel_SwapMainSub);
	//TAP_Channel_SwapMainSub(); 

	// swap main and sub channels and rescale to give the effect of a switch of sound
	TAP_Channel_GetCurrent(&tv_radio, &main_channel_number);
	TAP_Channel_GetCurrentSubTV(&sub_channel_number);
	if (debug) TAP_Print("Main channel %d, Sub channel %d\r\n", main_channel_number, sub_channel_number);
	TAP_Channel_Start(CHANNEL_Main, 0, sub_channel_number);
	TAP_Channel_Start(CHANNEL_Sub, 0, main_channel_number);
	TAP_Channel_Scale(CHANNEL_Main, sub_x, sub_y, sub_w, sub_h, TRUE);
	if (debug) TAP_Print("Scale main %d, %d, %d, %d\r\n", sub_x, sub_y, sub_w, sub_h);
	TAP_Channel_Scale(CHANNEL_Sub, main_x, main_y, main_w, main_h, TRUE);
	if (debug) TAP_Print("Scale sub %d, %d, %d, %d\r\n", main_x, main_y, main_w, main_h);
	
	// swap variables too
	channel_number = sub_channel_number;
	sub_channel_number = main_channel_number;
	main_channel_number = channel_number;
	x = sub_x, y = sub_y, w = sub_w, h = sub_h;
	sub_x = main_x, sub_y = main_y, sub_w = main_w, sub_h = main_h;
	main_x = x, main_y = y, main_w = w, main_h = h;
	scale_index = sub_scale_index;
	sub_scale_index = main_scale_index;
	main_scale_index = scale_index;

	handleInfoKey();

	if (debug) TAP_Print("PiP switched\r\n");
}

void handlePositionSwapKey()
{
	if (radio_mode)
		return;

	// swap variables
	x = sub_x, y = sub_y, w = sub_w, h = sub_h;
	sub_x = main_x, sub_y = main_y, sub_w = main_w, sub_h = main_h;
	main_x = x, main_y = y, main_w = w, main_h = h;
	scale_index = sub_scale_index;
	sub_scale_index = main_scale_index;
	main_scale_index = scale_index;

	// rescale main and sub channels 
	scaleMainAndSub();

	handleInfoKey();

	if (debug) TAP_Print("PiP swapped\r\n");
}

void handlePositionRevertKey()
{
	word i;

	if (recall_snap == SPLIT_SCREEN) {
		main_x = 10, main_y = 576/4, main_w = w_scales[6], main_h = h_scales[6];
		sub_x = 720/2+5, sub_y = 576/4, sub_w = w_scales[6], sub_h = h_scales[6];
		sub_scale_index = 6;
		main_scale_index = 6;
	}
	else {
		main_x = startup_main_x, main_y = startup_main_y, main_w = startup_main_w, main_h = startup_main_h;
		sub_x = startup_sub_x, sub_y = startup_sub_y, sub_w = startup_sub_w, sub_h = startup_sub_h;

		main_scale_index = -1; sub_scale_index = -1; 
		for (i = 0; i < NUMBER_OF_SCALES; i++)	{
			if (startup_main_w == w_scales[i]) main_scale_index = i;
			if (startup_sub_w == w_scales[i])  sub_scale_index = i;
		}
	}

	scaleMainAndSub();

	handleInfoKey();
}

void handleRadioModeKey()
{
	if (radio_playback)
		return;

	// abort if video playing
	Safe_TAP_Hdd_GetPlayInfo(&play_info);
	if (play_info.playMode == PLAYMODE_Playing || play_info.playMode == PLAYMODE_RecPlaying || play_info.playMode == PLAYMODE_TempPlaying)
		return;

	TAP_Channel_GetCurrent(&tv_radio, &main_channel_number);
	TAP_Channel_GetCurrentSubTV(&sub_channel_number);

	if (tv_radio == SVC_TYPE_Tv) {
		pre_radio_main_x = main_x, pre_radio_main_y = main_y, pre_radio_main_w = main_w, pre_radio_main_h = main_h;
		pre_radio_sub_x = sub_x, pre_radio_sub_y = sub_y, pre_radio_sub_w = sub_w, pre_radio_sub_h = sub_h;
		main_x = 720-250, main_y = 100, main_w = w_scales[NUMBER_OF_SCALES-1], main_h = h_scales[NUMBER_OF_SCALES-1];
		sub_x = 10, sub_y = 10, sub_w = w_scales[0], sub_h = h_scales[0];
		TAP_Channel_Start(CHANNEL_Main, SVC_TYPE_Radio, getRadioStartNumber());
		TAP_Channel_Start(CHANNEL_Sub, SVC_TYPE_Tv, main_channel_number);

		scaleMainAndSub();

		saved_main_number = reverse_pip ? sub_channel_number : main_channel_number;
		saved_pip_number = reverse_pip ? main_channel_number : sub_channel_number;
		radio_mode = TRUE;
	}
	else {
		main_x = pre_radio_main_x, main_y = pre_radio_main_y, main_w = pre_radio_main_w, main_h = pre_radio_main_h;
		sub_x = pre_radio_sub_x, sub_y = pre_radio_sub_y, sub_w = pre_radio_sub_w, sub_h = pre_radio_sub_h;
		TAP_Channel_Start(CHANNEL_Main, SVC_TYPE_Tv, reverse_pip ? saved_pip_number : saved_main_number);
		TAP_Channel_Start(CHANNEL_Sub, SVC_TYPE_Tv, reverse_pip ? saved_main_number : saved_pip_number);

		scaleMainAndSub();

		saved_radio_number = main_channel_number;
		radio_mode = FALSE;
	}

	handleInfoKey();
}

void enterToppyMode(dword delay)
{
	TAP_EnterNormal();
	toppy_mode_tick = TAP_GetTick();
	if (debug) TAP_Print("Pass control back to Toppy at %d\r\n", toppy_mode_tick);
	toppy_mode_delay = delay;
	toppy_mode = TRUE;
}

dword handleActivationKey(dword param1)
{
	if (param1 == pip_key || (pip_generated && isPipKey(param1)))
		return handlePipKey(param1, FALSE);

	if (reverse_pip_key && param1 == reverse_pip_key)
		return handlePipKey(param1, TRUE);

	return param1;
}

// RE checks (may wish to modify param1)
dword handleRemoteExtenderKeys(dword param1, dword param2)
{
	// RE checks
	switch (param2) {

		// channel up
		case RAWKEY_ChUp_Subt:

			changeChannelUp(TRUE);
			return 0;
				
		// channel down
		case RAWKEY_ChDown_Teletext:

			changeChannelDown(TRUE);
			return 0;

		// set list as being unique
		case RAWKEY_List_TVRad:

			return RAWKEY_List_TVRad;

		// volume up
		case RAWKEY_VolUp_Fav:

			return param1;

		// volume down
		case RAWKEY_VolDown_Exit:

			return param1;
	}

	return param1;
}

void handleHookEvent(dword param1, dword param2)
{
#ifndef WIN32
	// bdb code
	if (useFirmwareHacks()) {
		if (param1 == scale_hook_id) param1 = scale_hook(param1, param2);
		if (param1 == wss_hook_id) param1 = wss_hook(param1, param2);
	}
#endif
}

void handleInitialisation()
{
	if (debug) TAP_Print("Init\r\n"); 

	// load in all the configuration options
	if (debug) TAP_Print("Load config\r\n"); 
	LoadConfiguration();

	// default settings from config
	setDefaultSettings();

#ifndef WIN32
	firmware_hacks_available = TRUE;
	if (!firmware_hacks_enabled || exTAP_Init() < 0) {
		if (debug) TAP_Print("Firmware not supported for hackery\r\n");
		firmware_hacks_available = FALSE;
	}
#endif

	// get the number of tv/radio channels
	TAP_Channel_GetTotalNum(&number_of_tv_channels, &number_of_radio_channels);
	if (debug) TAP_Print("%d TV Channels, %d Radio Channels\r\n", number_of_tv_channels, number_of_radio_channels);

	// save the current channel
	TAP_Channel_GetCurrent(&tv_radio, &channel_number);
	if (tv_radio == SVC_TYPE_Tv)
		current_tv_channel_number = previous_tv_channel_number = channel_number;
	else
		current_radio_channel_number = previous_radio_channel_number = channel_number;

	// initialize the graphics engine
	setupGraphics();
}

void handleIdleEvent()
{
	if (handle_initialisation) {
		handleInitialisation();
		handle_initialisation = FALSE;
		return;
	}

	// do some stuff after the channel name has been drawn
	if (active && info_delay > 0
			&& (TAP_GetTick() > (channel_clear_name_tick + (info_delay * 100))) 
			&& (channel_clear_name_tick > 0)
			&& !(toppy_mode || external_mode || options_displayed)) {

		// refresh the border (this deletes the channel names)
		clearOsd();

		// to get round bug where main pictures snaps to full screen, rescale now
		if (!radio_mode)
			scaleMainAndSub();

		if (config_mode)
			drawConfigMode();
		drawCurrentWindowIcon();
		drawRadioIcon();

		channel_clear_name_tick = 0;
	}

	// blank out the config mode alert (only when turned off)
	if ((active) 
			&& (TAP_GetTick() > (config_mode_clear_tick + 200)) 
			&& (config_mode_clear_tick > 0)
			&& !(config_mode || toppy_mode || external_mode || options_displayed)) {

		// clear the Osd and reset
		clearOsd();
		drawCurrentWindowIcon();
		drawRadioIcon();
		config_mode_clear_tick = 0;

		if (info_delay == 0)
			handleInfoKey();
	}

	// blank out the saved config alert
	if ((active) 
			&& (TAP_GetTick() > (saved_config_clear_tick + 200)) 
			&& (saved_config_clear_tick > 0)
			&& !(toppy_mode || external_mode || options_displayed)) {

		// clear the Osd and reset
		clearOsd();

		if (config_mode)
			drawConfigMode();
		drawCurrentWindowIcon();
		drawRadioIcon();
		saved_config_clear_tick = 0;

		if (info_delay == 0)
			handleInfoKey();
	}

	// blank out the border
	if ((active) 
			&& (TAP_GetTick() > (border_clear_tick + 200)) 
			&& (border_clear_tick > 0)
			&& !(toppy_mode || external_mode || options_displayed || radio_mode)) {

		// clear the Osd and reset
		//if (main_x > 7 && main_y > 2 && (main_x + main_w + 10 + 4) < 720 && (main_y + main_h + 4) < 576)
		//	TAP_Osd_DrawRectangle(0, main_x-5, main_y, main_w+10, main_h, 2, COLOR_None);
		border_clear_tick = 0;

		if (info_delay == 0)
			handleInfoKey();
	}

	// if we have handed over control, wait till we can grab it back
	if (toppy_mode) {
		if (TAP_GetTick() > (toppy_mode_tick + toppy_mode_delay)) {
			TAP_GetState(&state, &sub_state);
			if (state == STATE_Normal && sub_state == SUBSTATE_Normal) {
				if (debug) TAP_Print("Returned from external after %d, %d\r\n", toppy_mode_tick, toppy_mode_delay);
				TAP_ExitNormal();
				TAP_Channel_GetCurrentSubTV(&sub_channel_number);
				TAP_Channel_Start(CHANNEL_Sub, 0, sub_channel_number);

				scaleMainAndSub();
				clearOsd();
				if (config_mode)
					drawConfigMode();
				drawCurrentWindowIcon();
				drawRadioIcon();
				drawChannelName(TRUE);
				drawChannelName(FALSE);
				
				toppy_mode = FALSE;
			}
		}
	}

	// check for channel change so we track the previous channel
	TAP_Channel_GetCurrent(&tv_radio, &channel_number);
	if (tv_radio == SVC_TYPE_Tv) {
		if (current_tv_channel_number == -1) {
			current_tv_channel_number = previous_tv_channel_number = channel_number;
		}
		ef (channel_number != current_tv_channel_number) {
			previous_tv_channel_number = current_tv_channel_number;
			current_tv_channel_number = channel_number;
		}
	}
	else {
		if (current_radio_channel_number == -1) {
			current_radio_channel_number = previous_radio_channel_number = channel_number;
		}
		ef (channel_number != current_radio_channel_number) {
			previous_radio_channel_number = current_radio_channel_number;
			current_radio_channel_number = channel_number;
		}
	}

#ifndef WIN32
	// bdb code
    //if (useFirmwareHacks()) if (aspect_enabled) exTAP_update_aspect_ratio();
#endif
}

dword handleNumberButtons(dword param1)
{
	switch (param1) {

		case RKEY_0:
		case RKEY_1:
		case RKEY_2:
		case RKEY_3:
		case RKEY_4:
		case RKEY_5:
		case RKEY_6:
		case RKEY_7:
		case RKEY_8:
		case RKEY_9:

			enterToppyMode(250);
			break;
	}

	return param1;
}

dword handleKeyChoiceEventKey(dword param1)
{
	dword selection = TAP_Win_GetSelection(&options_window);

	switch (param1) {

		case RKEY_Ok:

			if (edit_config_keys) {
				if (key_selection == 3 && keys[selection].key_value == RKEY_None)
					break;

				current_config_keys[key_selection].key_index = selection;
				config_keys[key_selection].key_value = keys[selection].key_value;
				options_changed = TRUE;
				key_choice_displayed = FALSE;
				drawConfigureKeysWindow(key_selection);
			}
			else {
				current_normal_keys[key_selection].key_index = selection;
				normal_keys[key_selection].key_value = keys[selection].key_value;
				options_changed = TRUE;
				key_choice_displayed = FALSE;
				drawConfigureKeysWindow(key_selection);
			}
			break;

		case RKEY_Exit:

			key_choice_displayed = FALSE;
			drawConfigureKeysWindow(key_selection);
			break;

		default:

			TAP_Win_Action(&options_window, param1);
			break;
	}

	return 0;
}

dword handleWindowConfigEventKey(dword param1)
{
	dword selection = TAP_Win_GetSelection(&options_window);

	switch (param1) {

		case RKEY_VolUp:

			switch (selection) {
				case 0: startup_main_x = (startup_main_x + startup_main_w + 1) > 720 ? 0 : startup_main_x + 1; break;
				case 1: startup_main_y = (startup_main_y + startup_main_h + 1) > 576 ? 0 : startup_main_y + 1; break;
				case 2: startup_main_w = (startup_main_x + startup_main_w + 1) > 720 ? 30 : startup_main_w + 1; break;
				case 3: startup_main_h = (startup_main_y + startup_main_h + 1) > 576 ? 30 : startup_main_h + 1; break;
				case 4: startup_sub_x = (startup_sub_x + startup_sub_w + 1) > 720 ? 0 : startup_sub_x + 1; break;
				case 5: startup_sub_y = (startup_sub_y + startup_sub_h + 1) > 576 ? 0 : startup_sub_y + 1; break;
				case 6: startup_sub_w = (startup_sub_x + startup_sub_w + 1) > 720 ? 30 : startup_sub_w + 1; break;
				case 7: startup_sub_h = (startup_sub_y + startup_sub_h + 1) > 576 ? 30 : startup_sub_h + 1; break;
			}
			main_x = startup_main_x, main_y = startup_main_y, main_w = startup_main_w, main_h = startup_main_h;
			sub_x = startup_sub_x, sub_y = startup_sub_y, sub_w = startup_sub_w, sub_h = startup_sub_h;
			scaleMainAndSub();
			options_changed = TRUE;
			TAP_Win_SetSelection(&options_window, selection);
			drawCurrentWindowSettings();
			break;

		case RKEY_VolDown:

			switch (selection) {
				case 0: startup_main_x = (startup_main_x == 0) ? 720 - startup_main_w : startup_main_x - 1; break;
				case 1: startup_main_y = (startup_main_y == 0) ? 576 - startup_main_h : startup_main_y - 1; break;
				case 2: startup_main_w = (startup_main_w <= 30) ? 720 - startup_main_x : startup_main_w - 1; break;
				case 3: startup_main_h = (startup_main_h <= 30) ? 576 - startup_main_y : startup_main_h - 1; break;
				case 4: startup_sub_x = (startup_sub_x == 0) ? 720 - startup_sub_w : startup_sub_x - 1; break;
				case 5: startup_sub_y = (startup_sub_y == 0) ? 576 - startup_sub_h : startup_sub_y - 1; break;
				case 6: startup_sub_w = (startup_sub_w <= 30) ? 720 - startup_sub_x : startup_sub_w - 1; break;
				case 7: startup_sub_h = (startup_sub_h <= 30) ? 576 - startup_sub_y : startup_sub_h - 1; break;
			}
			main_x = startup_main_x, main_y = startup_main_y, main_w = startup_main_w, main_h = startup_main_h;
			sub_x = startup_sub_x, sub_y = startup_sub_y, sub_w = startup_sub_w, sub_h = startup_sub_h;
			scaleMainAndSub();
			options_changed = TRUE;
			TAP_Win_SetSelection(&options_window, selection);
			drawCurrentWindowSettings();
			break;

		case RKEY_Forward:

			switch (selection) {
				case 0: startup_main_x = (startup_main_x + startup_main_w + 15) > 720 ? 0 : startup_main_x + 15; break;
				case 1: startup_main_y = (startup_main_y + startup_main_h + 15) > 576 ? 0 : startup_main_y + 15; break;
				case 2: startup_main_w = (startup_main_x + startup_main_w + 15) > 720 ? 30 : startup_main_w + 15; break;
				case 3: startup_main_h = (startup_main_y + startup_main_h + 15) > 576 ? 30 : startup_main_h + 15; break;
				case 4: startup_sub_x = (startup_sub_x + startup_sub_w + 15) > 720 ? 0 : startup_sub_x + 15; break;
				case 5: startup_sub_y = (startup_sub_y + startup_sub_h + 15) > 576 ? 0 : startup_sub_y + 15; break;
				case 6: startup_sub_w = (startup_sub_x + startup_sub_w + 15) > 720 ? 30 : startup_sub_w + 15; break;
				case 7: startup_sub_h = (startup_sub_y + startup_sub_h + 15) > 576 ? 30 : startup_sub_h + 15; break;
			}
			main_x = startup_main_x, main_y = startup_main_y, main_w = startup_main_w, main_h = startup_main_h;
			sub_x = startup_sub_x, sub_y = startup_sub_y, sub_w = startup_sub_w, sub_h = startup_sub_h;
			scaleMainAndSub();
			options_changed = TRUE;
			TAP_Win_SetSelection(&options_window, selection);
			drawCurrentWindowSettings();
			break;

		case RKEY_Rewind:

			switch (selection) {
				case 0: startup_main_x = (startup_main_x < 15) ? 720 - startup_main_w : startup_main_x - 15; break;
				case 1: startup_main_y = (startup_main_y < 15) ? 576 - startup_main_h : startup_main_y - 15; break;
				case 2: startup_main_w = (startup_main_w < 45) ? 720 - startup_main_x : startup_main_w - 15; break;
				case 3: startup_main_h = (startup_main_h < 45) ? 576 - startup_main_y : startup_main_h - 15; break;
				case 4: startup_sub_x = (startup_sub_x < 15) ? 720 - startup_sub_w : startup_sub_x - 15; break;
				case 5: startup_sub_y = (startup_sub_y < 15) ? 576 - startup_sub_h : startup_sub_y - 15; break;
				case 6: startup_sub_w = (startup_sub_w < 45) ? 720 - startup_sub_x : startup_sub_w - 15; break;
				case 7: startup_sub_h = (startup_sub_h < 45) ? 576 - startup_sub_y : startup_sub_h - 15; break;
			}
			main_x = startup_main_x, main_y = startup_main_y, main_w = startup_main_w, main_h = startup_main_h;
			sub_x = startup_sub_x, sub_y = startup_sub_y, sub_w = startup_sub_w, sub_h = startup_sub_h;
			scaleMainAndSub();
			options_changed = TRUE;
			TAP_Win_SetSelection(&options_window, selection);
			drawCurrentWindowSettings();
			break;

		case RKEY_Exit:

#ifndef WIN32
			TAP_Osd_SetTransparency(0, 255-toppy_transparency);
#endif
			window_config_displayed = FALSE;
			drawOptionsWindow(2);
			break;

		default:

			TAP_Win_Action(&options_window, param1);
			TAP_Win_SetSelection(&options_window, TAP_Win_GetSelection(&options_window));
			drawCurrentWindowSettings();
			break;
	}

	return 0;
}

dword handleKeyConfigEventKey(dword param1)
{
	dword selection = TAP_Win_GetSelection(&options_window);

	switch (param1) {

		case RKEY_Ok:

			key_selection = selection;
			key_choice_displayed = TRUE;
			drawChooseKeyWindow();
			break;

		case RKEY_Exit:

			key_config_displayed = FALSE;
			drawOptionsWindow(edit_config_keys ? 0 : 1);
			break;

		default:

			TAP_Win_Action(&options_window, param1);
			TAP_Win_SetSelection(&options_window, TAP_Win_GetSelection(&options_window));
			drawCurrentKeyValues();
			break;
	}

	return 0;
}

dword handleOptionsEventKey(dword param1)
{
	dword selection = TAP_Win_GetSelection(&options_window);

	switch (param1) {

		case RKEY_Ok:

			switch (selection) {

				case 0:
					key_config_displayed = TRUE;
					edit_config_keys = TRUE;
					drawConfigureKeysWindow(0);
					break;

				case 1:
					key_config_displayed = TRUE;
					edit_config_keys = FALSE;
					drawConfigureKeysWindow(0);
					break;

				case 2:
					toppy_transparency = TAP_GetSystemVar(SYSVAR_OsdAlpha);
#ifndef WIN32
					TAP_Osd_SetTransparency(0, 160);
#endif
					window_config_displayed = TRUE;
					drawWindowSettingsWindow();
					break;

				case 13:
					handleExitTap(active);
					break;
			}
			break;

		case RKEY_VolDown:

			switch (selection) {

				case 3:
					current_pip_index = current_pip_index == 0 ? KEY_COUNT - 1 : current_pip_index - 1;
					pip_key = keys[current_pip_index].key_value;
					while (pip_key == RKEY_None || pip_key == reverse_pip_key) {
						current_pip_index = current_pip_index == 0 ? KEY_COUNT - 1 : current_pip_index - 1;
						pip_key = keys[current_pip_index].key_value;
					}
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
					break;
				
				case 4:
					current_reverse_pip_index = current_reverse_pip_index == 0 ? KEY_COUNT - 1 : current_reverse_pip_index - 1;
					reverse_pip_key = keys[current_reverse_pip_index].key_value;
					while (pip_key == reverse_pip_key) {
						current_reverse_pip_index = current_reverse_pip_index == 0 ? KEY_COUNT - 1 : current_reverse_pip_index - 1;
						reverse_pip_key = keys[current_reverse_pip_index].key_value;
					}
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
					break;

				case 5:
					recall_snap = recall_snap ? FALSE : TRUE;
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
					break;

				case 6:
					topfield_model = topfield_model == 0 ? 2 : topfield_model - 1;
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
					break;

				case 7:
					masterpiece_mode = masterpiece_mode ? FALSE : TRUE;
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
					break;

				case 8:
					display_icons = display_icons == 0 ? 2 : display_icons - 1;
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
					break;

				case 9:
					info_delay = info_delay == 0 ? 60 : info_delay - 1;
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
					break;

				case 10:
		
#ifndef WIN32
					// bdb code
					if (useFirmwareHacks())
						close_exTras();
#endif
					firmware_hacks_enabled = firmware_hacks_enabled ? FALSE : TRUE;
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
#ifndef WIN32
					if (firmware_hacks_enabled) {
						firmware_hacks_available = TRUE;
						if (exTAP_Init() < 0) {
							if (debug) TAP_Print("Firmware not supported for hackery\r\n");
							firmware_hacks_available = FALSE;
						}
					}

					if (useFirmwareHacks())
						open_exTras();
#endif
					break;

				case 11:
					current_fav = current_fav < 0 ? fav_count - 1 : current_fav - 1;
					if (current_fav == -1)
						strcpy(fav_file_name, "");
					else
						strcpy(fav_file_name, favs[current_fav]);
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
					if (ChangeDirFavs())
						loadFavouriteChannels();
					ChangeDirSettings();
					break;

				case 12:
					current_skin = current_skin == 0 ? skin_count - 1 : current_skin - 1;
					strcpy(skin_file_name, skins[current_skin]);
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
					TAP_Hdd_ChangeDir("Skins");
					loadSkinElements();
					TAP_Hdd_ChangeDir("..");
					break;
			}
			break;

		case RKEY_VolUp:

			switch (selection) {

			    case 3:
					current_pip_index = (current_pip_index + 1) % KEY_COUNT;					
					pip_key = keys[current_pip_index].key_value;
					while (pip_key == RKEY_None || pip_key == reverse_pip_key) {
						current_pip_index = (current_pip_index + 1) % KEY_COUNT ;
						pip_key = keys[current_pip_index].key_value;
					}
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
					break;
				
				case 4:
					current_reverse_pip_index = (current_reverse_pip_index + 1) % KEY_COUNT;					
					reverse_pip_key = keys[current_reverse_pip_index].key_value;
					while (pip_key == reverse_pip_key) {
						current_reverse_pip_index = (current_reverse_pip_index + 1) % KEY_COUNT;		
						reverse_pip_key = keys[current_reverse_pip_index].key_value;
					}
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
					break;

				case 5:
					recall_snap = recall_snap ? FALSE : TRUE;
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
					break;

				case 6:
					topfield_model = (topfield_model + 1) % 3;
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
					break;

				case 7:
					masterpiece_mode = masterpiece_mode ? FALSE : TRUE;
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
					break;

				case 8:
					display_icons = (display_icons + 1) % 3;
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
					break;

				case 9:
					info_delay = (info_delay + 1) % 61;
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
					break;

				case 10:
#ifndef WIN32
					// bdb code
					if (useFirmwareHacks())
						close_exTras();
#endif
					firmware_hacks_enabled = firmware_hacks_enabled ? FALSE : TRUE;
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
#ifndef WIN32
					if (firmware_hacks_enabled) {
						firmware_hacks_available = TRUE;
						if (exTAP_Init() < 0) {
							if (debug) TAP_Print("Firmware not supported for hackery\r\n");
							firmware_hacks_available = FALSE;
						}
					}

					if (useFirmwareHacks())
						open_exTras();
#endif
					break;

				case 11:
					current_fav = current_fav == (fav_count - 1) ? -1 : current_fav + 1;
					if (current_fav == -1)
						strcpy(fav_file_name, "");
					else
						strcpy(fav_file_name, favs[current_fav]);
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
					if (ChangeDirFavs())
						loadFavouriteChannels();
					ChangeDirSettings();
					break;

				case 12:
					current_skin = (current_skin + 1) % skin_count;
					strcpy(skin_file_name, skins[current_skin]);
					options_changed = TRUE;
					TAP_Win_SetSelection(&options_window, selection);
					drawOptionValues();
					TAP_Hdd_ChangeDir("Skins");
					loadSkinElements();
					TAP_Hdd_ChangeDir("..");
					break;
			}
			break;

		case RKEY_Exit:
		case RKEY_Menu:

			SaveConfigurationToFile();
			TAP_Win_Delete(&options_window);
			
			if (active) {
				clearOsd();
				if (options_changed)
					drawSavedConfig();
				else {
					if (config_mode)
						drawConfigMode();
					drawCurrentWindowIcon();
					drawRadioIcon();
				}
			}
			else
				TAP_EnterNormal();

			options_displayed = FALSE;
			break;

		default:

			TAP_Win_Action(&options_window, param1);
			TAP_Win_SetSelection(&options_window, TAP_Win_GetSelection(&options_window));
			drawOptionValues();
			break;
	}

	return 0;
}

dword handleOptionMenusEventKeys(dword param1)
{
	if (window_config_displayed)
		return handleWindowConfigEventKey(param1);
	ef (key_choice_displayed)
		return handleKeyChoiceEventKey(param1);
	ef (key_config_displayed)
		return handleKeyConfigEventKey(param1);
	else
		return handleOptionsEventKey(param1);
}

dword handleCustomKeys(dword param1, dword param3, bool custom_config_keys)
{
	word i;

	if (custom_config_keys) {
		for (i = 0; i < CURRENT_CONFIG_KEY_COUNT; i++) {
			if (getCurrentKeyValue(i, TRUE) == param3) {
				switch (i) {

					case 1:

						handlePositionRevertKey();
						return 0;

					case 2: 

						handlePositionSwapKey();
						return 0;

					case 3:

						if (!radio_mode) {
							saveCurrentSettings();
							drawSavedConfig();
						}
						return 0;

					case 4:

						shiftChannelUp(TRUE);
						return 0;

					case 5:

						shiftChannelDown(TRUE);
						return 0;

					case 6:

						shiftChannelLeft(TRUE);
						return 0;

					case 7:

						shiftChannelRight(TRUE);
						return 0;

					case 8:

						shrinkChannel(TRUE);
						return 0;

					case 9:

						enlargeChannel(TRUE);
						return 0;

					case 10:

						shiftChannelUp(FALSE);
						return 0;

					case 11:

						shiftChannelDown(FALSE);
						return 0;

					case 12:

						shiftChannelLeft(FALSE);
						return 0;

					case 13:

						shiftChannelRight(FALSE);
						return 0;

					case 14:

						shrinkChannel(FALSE);
						return 0;

					case 15:

						enlargeChannel(FALSE);
						return 0;

					case 16:

						changeChannelUp(FALSE);
						return 0;

					case 17:

						changeChannelDown(FALSE);
						return 0;

					case 18:

						handleClosePip(FALSE);
						return 0;
				}
			}
		}
	}
	else {
		for (i = 0; i < CURRENT_NORMAL_KEY_COUNT; i++) {
			if (getCurrentKeyValue(i, FALSE) == param3) {
				switch (i) {
			
					case 0:

						changeChannelUp(TRUE);
						return 0;

					case 1:

						changeChannelDown(TRUE);
						return 0;

					case 2:

						changeChannelUp(FALSE);
						return 0;

					case 3:

						changeChannelDown(FALSE);
						return 0;

					case 4:

						handlePositionRevertKey();
						return 0;

					case 5: 

						handlePositionSwapKey();
						return 0;

					case 6:

						if (!radio_mode) {
							saveCurrentSettings();
							drawSavedConfig();
						}
						return 0;

					case 7:

						shiftChannelUp(TRUE);
						return 0;

					case 8:

						shiftChannelDown(TRUE);
						return 0;

					case 9:

						shiftChannelLeft(TRUE);
						return 0;

					case 10:

						shiftChannelRight(TRUE);
						return 0;

					case 11:

						shrinkChannel(TRUE);
						return 0;

					case 12:

						enlargeChannel(TRUE);
						return 0;

					case 13:

						shiftChannelUp(FALSE);
						return 0;

					case 14:

						shiftChannelDown(FALSE);
						return 0;

					case 15:

						shiftChannelLeft(FALSE);
						return 0;

					case 16:

						shiftChannelRight(FALSE);
						return 0;

					case 17:

						shrinkChannel(FALSE);
						return 0;

					case 18:

						enlargeChannel(FALSE);
						return 0;

					case 19:

						handleClosePip(reverse_pip ? FALSE : TRUE);
						return 0;
				}
			}
		}
	}

	return param1;
}

dword handleNormalModeKeys(dword param1, dword param3)
{
	dword custom_handled;

	if (debug) TAP_Print("SP Normal mode received key %d\r\n", param3);

	// first check the customised keys to see if one has been pressed
	custom_handled = handleCustomKeys(param1, param3, FALSE);
	if (custom_handled == 0)
		return custom_handled;

	switch (param3) {

		// if special super key press, then show the super pip interface
		case Super_PiP:

			handleSuperPipKey();
			return 0;

		// display the options menu
		case RKEY_Menu:

			ChangeDirSettings();
			drawOptionsWindow(0);
			options_displayed = TRUE;
			options_changed = FALSE;
			if (debug) TAP_Print("Created window\r\n");
			return 0;

		// switch current window
		case RKEY_PipSwap:

			handlePipSwapKey();
			return 0;

		// switch to radio mode
		case RKEY_TvRadio:

			handleRadioModeKey();
			return 0;

		// default following keys to standard behaviour
		case RKEY_Filelist:
		case RKEY_Guide:
		case RKEY_Slow:
		case RKEY_Record:
		case RKEY_Pause:
		case RKEY_Rewind:
		case RKEY_Forward:
		case RKEY_Recall:

		case RKEY_White:
		case RKEY_Ok:
		case RKEY_Play:
		case RKEY_Stop:
		case RKEY_Options:
		case RKEY_Text:
		case RAWKEY_List_TVRad:

		case RKEY_Red:
		case RKEY_Blue:
		case RKEY_Yellow:
		case RKEY_Green:

			// abort skip keys if not in playback mode
			if (param1 == RKEY_Red || param1 == RKEY_Blue || param1 == RKEY_Yellow || param1 == RKEY_Green) {
				Safe_TAP_Hdd_GetPlayInfo(&play_info);
				if (play_info.playMode != PLAYMODE_Playing && play_info.playMode != PLAYMODE_RecPlaying && play_info.playMode != PLAYMODE_TempPlaying)
					return 0;
			}

			// abort archive or back if in radio mode
			if ((param1 == RKEY_Filelist || param1 == RKEY_Recall) && radio_mode)
				return 0;

			TAP_GenerateEvent(EVT_KEY, param1, 0);
			enterToppyMode(50);
			return 0;

		// display the channel details
		case RKEY_Info:

			handleInfoKey();
			return 0;
		
		// change current channel up
		/*case RKEY_ChUp:

			changeChannelUp(TRUE);
			return 0;

		// change current channel down
		case RKEY_ChDown:

			changeChannelDown(TRUE);
			return 0;

		// change sub channel up
		case RKEY_VolUp:

			changeChannelUp(FALSE);
			return 0;

		// change sub channel down
		case RKEY_VolDown:

			changeChannelDown(FALSE);
			return 0;
		*/

		// exit pip
		case RKEY_Exit:

			handleClosePip(reverse_pip ? FALSE : TRUE);
			return 0;

		// handle numbers for channel changing
		default:

			param1 = handleNumberButtons(param1);
			break;
	}

	return param1;
}

dword handleConfigModeKeys(dword param1, dword param3)
{
	dword custom_handled;

	// first check the customised keys to see if one has been pressed
	custom_handled = handleCustomKeys(param1, param3, TRUE);
	if (custom_handled == 0)
		return custom_handled;

	// if no custom key pressed, then handle standard functions
	switch (param3) {

		// display the options menu
		case RKEY_Menu:

			ChangeDirSettings();
			drawOptionsWindow(0);
			options_displayed = TRUE;
			options_changed = FALSE;
			if (debug) TAP_Print("Created window\r\n");
			return 0;

		// switch current window
		case RKEY_PipSwap:

			handlePipSwapKey();
			return 0;

		// change current channel via listing
		case RKEY_TvRadio:

			handleRadioModeKey();
			return 0;

		// default following keys to standard behaviour
		case RKEY_Filelist:
		case RKEY_Guide:
		case RKEY_Slow:
		case RKEY_Record:
		case RKEY_Pause:
		case RKEY_Rewind:
		case RKEY_Forward:
		case RKEY_Recall:

		case RKEY_White:
		case RKEY_Ok:
		case RKEY_Play:
		case RKEY_Stop:
		case RKEY_Options:
		case RKEY_Text:
		case RAWKEY_List_TVRad:

		case RKEY_Red:
		case RKEY_Blue:
		case RKEY_Yellow:
		case RKEY_Green:

			// abort skip keys if not in playback mode
			if (param1 == RKEY_Red || param1 == RKEY_Blue || param1 == RKEY_Yellow || param1 == RKEY_Green) {
				Safe_TAP_Hdd_GetPlayInfo(&play_info);
				if (play_info.playMode != PLAYMODE_Playing && play_info.playMode != PLAYMODE_RecPlaying && play_info.playMode != PLAYMODE_TempPlaying)
					return 0;
			}

			// abort archive or back if in radio mode
			if ((param1 == RKEY_Filelist || param1 == RKEY_Recall) && radio_mode)
				return 0;

			TAP_GenerateEvent(EVT_KEY, param1, 0);
			enterToppyMode(50);
			return 0;

		// display the channel details
		case RKEY_Info:

			handleInfoKey();
			return 0;

		// exit pip
		case RKEY_Exit:

			handleClosePip(reverse_pip ? FALSE : TRUE);
			return 0;

		// handle the number buttons for channel changing
		default:

			param1 = handleNumberButtons(param1);
			break;
	}

	return param1;
}

dword handleActiveKeys(dword param1, dword param3)
{
	// handle config mode first
	if (param3 == getCurrentKeyValue(0, TRUE)) {
		config_mode = config_mode ? FALSE : TRUE;
		drawConfigMode();
		//drawCurrentWindowIcon();
		//drawRadioIcon();
		return 0;
	}

	if (isPipKey(param3))
		return handlePipKey(param3, FALSE);

	// handle config and normal keys separately
	return config_mode ? handleConfigModeKeys(param1, param3) : handleNormalModeKeys(param1, param3);
}

#define P2RKEY_TvRadio  0x1000e
#define P2RKEY_Left     0x1000f
#define P2RKEY_Up       0x10012
#define P2RKEY_Right    0x10015
#define P2RKEY_VolDown  0x10017
#define P2RKEY_List     0x10018
#define P2RKEY_Down     0x1001d
#define P2RKEY_VolUp    0x10020
#define P2RKEY_ChUp     0x10021
#define P2RKEY_ChDown   0x10037

dword TAP_EventHandler(word event_type, dword param1, dword param2)
{
	if (event_type == EVT_KEY)
    {
		dword param3;

		//if (debug) TAP_Print("SP Received key %d\r\n", param1);

		// check for pip key if not active
		if (!active && !options_displayed) {
			return handleActivationKey(param1);
		}

		// if in toppy_mode, then may want to accept another number button press
		if (toppy_mode)
			return handleNumberButtons(param1);

		// handle the various option menus
		if (options_displayed) {
			return handleOptionMenusEventKeys(param1);
		}

		// RE checks (may wish to modify param1)
		//re_result = handleRemoteExtenderKeys(&param1, param2);
		//if (re_result >= 0)
		//	return re_result;
		// RE checks
		switch (param2) {

			// channel up
			case RAWKEY_ChUp_Subt:
			case P2RKEY_ChUp:

				changeChannelUp(TRUE);
				return 0;
					
			// channel down
			case RAWKEY_ChDown_Teletext:
			case P2RKEY_ChDown:

				changeChannelDown(TRUE);
				return 0;

			// set list as being unique
			case RAWKEY_List_TVRad:
			case P2RKEY_List:

				param3 = RAWKEY_List_TVRad;
				break;

			// volume up
			case RAWKEY_VolUp_Fav:
			case P2RKEY_VolUp:

				return param1;

			// volume down
			case RAWKEY_VolDown_Exit:
			case P2RKEY_VolDown:

				return param1;

			// everything else gets put in param3
			default:

				param3 = param1;
				break;
		}

		// finally handle superpip actions
		return handleActiveKeys(param1, param3);
	}

	if (event_type == EVT_IDLE)
		handleIdleEvent();

#ifndef WIN32
	if (event_type == EVT_HOOK)
		handleHookEvent(param1, param2);
#endif

    return param1;
}

int TAP_Main(void)
{
	if (debug) TAP_Print("\rStarting TAP in TSR mode\r\n");

	handle_initialisation = TRUE;
	
	// TSR mode
#ifndef WIN32
	return 2;
#else
    return 1;
#endif
}

//----------------------------------------------------------------------------
// This function cleans up and closes the TAP
bool TSRCommanderExitTAP()
{
	handleExitTap(FALSE);

	return TRUE;
}


//-----------------------------------------------------------------------------
// This function displays a config menu
void TSRCommanderConfigDialog()
{
	ChangeDirSettings();
	TAP_ExitNormal();
	drawOptionsWindow(0);
	options_displayed = TRUE;
	options_changed = FALSE;
}

#ifdef WIN32
int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPTSTR    lpCmdLine,
                     int       nCmdShow)
{
   // The following line sets some information in the SDK that will be available through the File->Info menuchoice
   setTAP_Information(__tap_program_name__, __tap_author_name__, __tap_description__, __tap_etc_str__, __tap_ud__);

   enableFramebufferUpdate(FALSE); // Set to TRUE to get all OSD output copied to framebuffer.
                                   // (available through GetBaseInfo). Setting this to TRUE will slow things down, so don't
                                   // use if you don't need it.

   enableFramebufferMonitor(FALSE); // Set to TRUE to get the SDK to copy any changed contents of the
                                    // framebuffer to the emulated screen. 

   return InitializeTAP_Environment(hInstance, nCmdShow, __tap_program_name__, TAP_Main, TAP_EventHandler);
} 
#endif