/*
GfxLIB - A utility library to assist in doing graphics on the OSD on Topfield 5000.
Author : Kjetil Naess, July 2005
Version: 1.0

You may copy and/or use GfxLIB in either source or compiled form as you wish. Any suggestions and/or
improvements are welcome so we can make this a good public resource.
*/

#ifdef WIN32
//#pragma comment(lib, "c:\\TopfieldDisk\\DataFiles\\SDK\\TAP_SDK_DLL")
#include "TAP_Emulations.h"
#endif

#include "tap.h"
#include "gfxlib.h"

word *_vAddr[VideoResY];

byte currentROP = blt_SRCCOPY;

void setupGraphics()
{
    int i;

    TYPE_OsdBaseInfo osdBaseInfo;
    TAP_Osd_GetBaseInfo( &osdBaseInfo );

    // set vertical address
    for( i=0; i<VideoResY; i+=2 )
    {
	    _vAddr[i] = (word*)osdBaseInfo.eAddr + 360*(i+0) + 0;
	    _vAddr[i+1] = (word*)osdBaseInfo.oAddr + 360*(i+0) + 0;
    }
}

void endGraphics()
{
}
  
void setPixel(int x, int y, word color)
{
   //HDC dc;      
   if (x <= xLimit && y <= yLimit && (x >= 0 && y >= 0))   
   {
  	  *(_vAddr[y]+x) = color;
   }
}

word getPixel(int x, int y)
{
   return *( _vAddr[y]+x);
}

//Bresenham circle with center at (xc,yc) with radius and red green blue color
bool drawCircle(int xc, int yc, int radius, word color)
{
    int x = 0;
    int y = radius;
    int p = 3 - (radius << 1);
    int a, b, c, d, e, f, g, h;
    if(xc - radius < 0 || xc + radius >= VideoResX || yc - radius < 0 || yc + radius >= VideoResY) return 0;
    while (x <= y)
    {
         a = xc + x; //8 pixels can be calculated at once thanks to the symmetry
         b = yc + y;
         c = xc - x;
         d = yc - y;
         e = xc + y;
         f = yc + x;
         g = xc - y;
         h = yc - x;
         setPixel(a, b, color);
         setPixel(c, d, color);
         setPixel(e, f, color);
         setPixel(g, f, color);
         if(x>0) //avoid drawing pixels at same position as the other ones
         {
             setPixel(a, d, color);
             setPixel(c, b, color);
             setPixel(e, h, color);
             setPixel(g, h, color);
         }
         if(p < 0) p += (x++ << 2) + 6;
         else p += ((x++ - y--) << 2) + 10;
  }
  
  return 1;
}

//Filled bresenham circle with center at (xc,yc) with radius and red green blue color
void drawDisk(int xc, int yc, int radius, word color)
{
    int x = 0;
    int y = radius;
    int p = 3 - (radius << 1);
    int a, b, c, d, e, f, g, h;
    int pb = 0;
	int pd = 0; //previous values: to avoid drawing horizontal lines multiple times    

	if(xc + radius < 0 || xc - radius >= VideoResX || yc + radius < 0 || yc - radius >= VideoResY) return ; //every single pixel outside screen, so don't waste time on it
    while (x <= y)
    {
         // write data
         a = xc + x;
         b = yc + y;
         c = xc - x;
         d = yc - y;
         e = xc + y;
         f = yc + x;
         g = xc - y;
         h = yc - x;
         if(b != pb) drawLine(a, b, c, b, color);
         if(d != pd) drawLine(a, d, c, d, color);
         if(f != b)  drawLine(e, f, g, f, color);
         if(h != d && h != f) drawLine(e, h, g, h,color);
         pb = b;
         pd = d;         
         if(p < 0) p += (x++ << 2) + 6;
         else p += ((x++ - y--) << 2) + 10;
  }
}

void drawLine(int x1, int y1, int x2, int y2, word color)
{
int  x, y;
 int  dx, dy;
 int  incx, incy;
 int  balance;
 
 if (x2 >= x1)
 {
    dx = x2 - x1;
    incx = 1;
 }
 else
 {
    dx = x1 - x2;
    incx = -1;
 }

 if (y2 >= y1)
 {
    dy = y2 - y1;
    incy = 1;
 }
 else
 {
    dy = y1 - y2;
    incy = -1;
 }

 x = x1;
 y = y1;

 if (dx >= dy)
 {
  dy <<= 1;
  balance = dy - dx;
  dx <<= 1;

  while (x != x2)
  {
   setPixel(x, y,color);
   if (balance >= 0)
   {
    y += incy;
    balance -= dx;
   }
   balance += dy;
   x += incx;
  } setPixel(x, y,color);
 }
 else
 {
  dx <<= 1;
  balance = dx - dy;
  dy <<= 1;

  while (y != y2)
  {
   setPixel(x, y,color);
   if (balance >= 0)
   {
    x += incx;
    balance -= dy;
   }
   balance += dx;
   y += incy;
  } setPixel(x, y,color);
 }
}

void DuffsDevice(register unsigned char *to, register unsigned char *from, register unsigned int count)
{
   register unsigned int n = (count + 7) / 8;
   switch(count%8)
   {
   case 0: do
           {
           *to++ = *from++;
   case 7: *to++ = *from++;
   case 6: *to++ = *from++;
   case 5: *to++ = *from++;
   case 4: *to++ = *from++;
   case 3: *to++ = *from++;
   case 2: *to++ = *from++;
   case 1: *to++ = *from++;
           } while (--n > 0);
   }  
}
void DuffsDeviceA(register unsigned char *to, register unsigned char *from, register unsigned int count)
{
   register unsigned int n = (count + 7) / 8;
   switch(count%8)
   {
   case 0: do
           {
           *to++ &= *from++;
   case 7: *to++ &= *from++;
   case 6: *to++ &= *from++;
   case 5: *to++ &= *from++;
   case 4: *to++ &= *from++;
   case 3: *to++ &= *from++;
   case 2: *to++ &= *from++;
   case 1: *to++ &= *from++;
           } while (--n > 0);
   }  
}
void DuffsDeviceX(register unsigned char *to, register unsigned char *from, register unsigned int count)
{
   register unsigned int n = (count + 7) / 8;
   switch(count%8)
   {
   case 0: do
           {
           *to++ ^= *from++;
   case 7: *to++ ^= *from++;
   case 6: *to++ ^= *from++;
   case 5: *to++ ^= *from++;
   case 4: *to++ ^= *from++;
   case 3: *to++ ^= *from++;
   case 2: *to++ ^= *from++;
   case 1: *to++ ^= *from++;
           } while (--n > 0);
   }  
}
void DuffsDeviceO(register unsigned char *to, register unsigned char *from, register unsigned int count)
{
   register unsigned int n = (count + 7) / 8;
   switch(count%8)
   {
   case 0: do
           {
           *to++ |= *from++;
   case 7: *to++ |= *from++;
   case 6: *to++ |= *from++;
   case 5: *to++ |= *from++;
   case 4: *to++ |= *from++;
   case 3: *to++ |= *from++;
   case 2: *to++ |= *from++;
   case 1: *to++ |= *from++;
           } while (--n > 0);
   }  
}

void DuffsDeviceF(register unsigned char *to, register unsigned char *from, register unsigned int count)
{
   register unsigned int n = (count + 7) / 8;
   switch(count%8)
   {
   case 0: do
           {
           *to++ = *from;
   case 7: *to++ = *from;
   case 6: *to++ = *from;
   case 5: *to++ = *from;
   case 4: *to++ = *from;
   case 3: *to++ = *from;
   case 2: *to++ = *from;
   case 1: *to++ = *from;
           } while (--n > 0);
   }  
}

void drawHorizLine(int x1, int y, int w, word color)
{
   int i;
   for (i = x1; i < x1+w; i++)
   {
       setPixel(i, y, color);
   }
}

void drawArea(int x1, int y1, int w, int h, word color)
{
   int i;
   for (i = y1; i < y1+h; i++)
	   drawHorizLine(x1, i, w, color);
}

void drawRectangle(int x1, int y1, int x2, int y2, word color, byte Filled)
{
   if (Filled == FALSE)
   {
      drawLine(x1, y1, x2, y1, color);
	  drawLine(x1, y1, x1, y2, color);
	  drawLine(x1, y2, x2, y2, color);
	  drawLine(x2, y2, x2, y1, color);
   }
   else
   {
      drawArea(x1, y1, x2-x1, y2-y1, color);
   }
}

void bitBlt(int x, int y, int w, int h, word *Data, unsigned char  Direction, unsigned char Operation)
{
   int i;
   int width=sizeof(word) * w;
   unsigned char *Destination;
   unsigned char *Source;
   for (i =0; i < h; i++)
   {
	   if (Direction == blt_ToScreen)
	   {
	      Destination = (unsigned char *)(_vAddr[i+y]+x);
		  Source = (unsigned char *)Data + (i * width);
	   }
	   else
	   {
	      Source = (unsigned char *)(_vAddr[i+y]+x);
		  Destination = (unsigned char *)Data + (i * width);
	   }
	   switch(Operation)
	   {
	   case blt_SRCCOPY : DuffsDevice((unsigned char *)Destination, (unsigned char *)Source, width); break;
	   case blt_SRCAND  : DuffsDeviceA((unsigned char *)Destination, (unsigned char *)Source, width); break;
	   case blt_SRCINVERT:DuffsDeviceX((unsigned char *)Destination, (unsigned char *)Source, width); break;
	   case blt_SRCPAINT: DuffsDeviceO((unsigned char *)Destination, (unsigned char *)Source, width); break;
	   }
   }
}

/*void drawStringCentre(int x1, int y1, int w, char *s)
{
   int xo, xc;

   xc = (w / 2) + x1;
   xo = (strlen(s) * 16) / 2;

   drawString(xc - xo,y1,s);
}*/

byte setROP(byte ROP)
{
   byte b = currentROP;
   currentROP = ROP;
   return b;
}

/*
 * This function will draw a quadrant of a circle. By using the Bresenham algorithm which
 * devides a circle into octets, this is easily achived.
 * quadrant=1 UpperLeft
 * quadrant=2 UpperRight
 * quadrant=3 LowerLeft
 * quadrant=4 LowerRight
 */
void gxCircleQuadrant(int xc, int yc, int radius, word color, int quadrant, byte Filled)
{
    int x = 0;
    int y = radius;
    int p = 3 - (radius << 1);
    int a, b, c, d, e, f, g, h;
    
	while (x <= y)
    {
         a = xc + x; //8 pixels can be calculated at once thanks to the symmetry
         b = yc + y;
         c = xc - x;
         d = yc - y;
         e = xc + y;
         f = yc + x;
         g = xc - y;
         h = yc - x;
         if (quadrant == 4) 
		 {
			 if (Filled)
				 drawLine(a, b, xc, b, color);
		     else
				 setPixel(a, b, color);
		 }
         if (quadrant == 1) 
		 {
			 if (Filled)
				 drawLine(c, d, xc,d,color);
			 else
				setPixel(c, d, color);
		 }
		 if (quadrant == 4) 
		 {
			 if (Filled)
				 drawLine(e, f, xc, f, color);
		     else
				 setPixel(e, f, color);
		 }
         if (quadrant == 3) 
		 {
			 if (Filled)
				 drawLine(g, f, xc, f, color);
		     else
				 setPixel(g, f, color);
		 }
         if(x>0) //avoid drawing pixels at same position as the other ones
         {
             if (quadrant == 2) 
			 {
				 if (Filled)
					 drawLine(a, d, xc, d, color);
			     else
					 setPixel(a, d, color);
			 }
             if (quadrant == 3) 
			 {
				 if (Filled)
					 drawLine(c, b, xc, b, color);
			     else
					 setPixel(c, b, color);
			 }
             if (quadrant == 2) 
			 {
				 if (Filled)
					 drawLine(e, h, xc, h, color);
			     else
					 setPixel(e, h, color);
			 }
             if (quadrant == 1) 
			 {
				 if (Filled)
					drawLine(g, h, xc,h, color);
			     else
   				    setPixel(g, h, color);
			 }
         }

		 if (p < 0) 
            p += (x++ << 2) + 6;
         else 
            p += ((x++ - y--) << 2) + 10;
	}

	/*while (x <= y)
    {
         a = xc + x; //8 pixels can be calculated at once thanks to the symmetry
         b = yc + y;
         c = xc - x;
         d = yc - y;
         e = xc + y;
         f = yc + x;
         g = xc - y;
         h = yc - x;
         if (quadrant == 4)  setPixel(a, b, color);
         if (quadrant == 1) setPixel(c, d, color);
         if (quadrant == 4) setPixel(e, f, color);
         if (quadrant == 3) setPixel(g, f, color);
         if(x>0) //avoid drawing pixels at same position as the other ones
         {
             if (quadrant == 2) setPixel(a, d, color);
             if (quadrant == 3) setPixel(c, b, color);
             if (quadrant == 2) setPixel(e, h, color);
             if (quadrant == 1) setPixel(g, h, color);
         }
         if (p < 0) 
            p += (x++ << 2) + 6;
         else 
            p += ((x++ - y--) << 2) + 10;
  }*/
}

#define stackSize 16384
static long stack[stackSize];
static long stackPointer = 0;

#define maxheight  600
#define maxwidth 800

byte pop(int *x, int *y) 
{ 
    if(stackPointer > 0) 
    { 
        int p = stack[stackPointer]; 
        *x = p / maxheight; 
        *y = p % maxheight; 
        stackPointer--; 
        return TRUE; 
    }     
    else 
    { 
        return FALSE; 
    }    
}    

byte push(int x, int y) 
{ 
    if(stackPointer < stackSize - 1) 
    { 
        stackPointer++; 
        stack[stackPointer] = maxheight * x + y; 
        return TRUE; 
    }     
    else 
    { 
        return FALSE; 
    }    
}     

void emptyStack() 
{ 
    int x, y; 
    while(pop(&x, &y)); 
}
/*
 * This implements a stackbased scanline floodfill. newColor is the color to check against, ie.
 * the boundries. oldColor is the color to be replaced. So, for a red shape on an empty
 * background call gxFloodFill(x, y, COLOR_Red, COLOR_None);
 */
void gxFloodFill(int x, int y, word newColor, word oldColor)
{
    int y1; 
    byte spanLeft, spanRight;

	if(oldColor == newColor) return;
    emptyStack();
    
    
    if(!push(x, y)) return;
    
    while(pop(&x, &y))
    {    
        y1 = y;
        while (getPixel(x, y1) == oldColor && y1 >= 0) 
			   y1--;
        y1++;
        spanLeft = spanRight = 0;
        while(getPixel(x, y1) == oldColor && y1 < maxheight)
        {
            setPixel(x, y1, newColor);
            if(!spanLeft && x > 0 && getPixel(x - 1, y1) == oldColor) 
            {
                if(!push(x - 1, y1)) return;
                spanLeft = 1;
            }
            else if(spanLeft && x > 0 && getPixel(x - 1, y1) != oldColor)
            {
                spanLeft = 0;
            }
            if(!spanRight && x < maxwidth - 1 && getPixel(x + 1, y1) == oldColor) 
            {
                if(!push(x + 1, y1)) return;
                spanRight = 1;
            }
            else if(spanRight && x < maxwidth - 1 && getPixel(x + 1, y1) != oldColor)
            {
                spanRight = 0;
            } 
            y1++;                   
        }
    }        
}

/*
 * Draws a rectangle with rounded edges. (x1,x2)-(x2-y2) is the extend of the rectangle,
 * r is the radius to draw the rounded corners with, color is color and if Filled==TRUE
 * it is drawn filled 
 */
void gxRoundRectangle(int x1, int y1, int x2, int y2, word color, byte Filled, int r)
{
   int X1, X2, X3, Y1, Y2, Y3;

   drawLine(x1,y1+r,x1,y2-r,color);
   drawLine(x1+r,y1,x2-r,y1,color);
   drawLine(x2,y1+r,x2,y2-r,color);
   drawLine(x1+r,y2,x2-r,y2,color);

   gxCircleQuadrant(x1+r,y1+r,r,color, 1, FALSE); // Upper left corner
   gxCircleQuadrant(x2-r,y1+r,r,color,2, FALSE);  // Upper right corner
   gxCircleQuadrant(x1+r,y2-r,r,color,3, FALSE);  // Lower left corner
   gxCircleQuadrant(x2-r,y2-r,r,color,4, FALSE);  // Lower right corner
  
   if (Filled) {
      //gxFloodFill(x1+r,y1+r, color, COLOR_None);
	    drawArea(x1,y1+r,x2-x1,(y2-y1)-(r<<1),color); // Draw center block as filled area
		drawArea(x1+r,y1,(x2-x1)-(r<<1),r, color);    // Top rectangle
		drawArea(x1+r,y2-r,(x2-x1)-(r<<1),r, color);  // Bottom rectangle
   }
}

void drawRoundedRectangle(word x, word y, word w, word h, word border, word border_colour, word fill_colour)
{
#ifndef WIN32
	int i, j;

	// draw the outer rounded rectangle
	gxRoundRectangle(x-5-border, y-5-border, x+w+5+border, y+h+5+border, border_colour, FALSE, 10+border);

	// draw the inner rounded rectangle
	if (fill_colour != COLOR_None)
		gxRoundRectangle(x-5, y-5, x+w+5, y+h+5, fill_colour, TRUE, 10);
	else
		gxRoundRectangle(x-5, y-5, x+w+5, y+h+5, border_colour, FALSE, 10);

	// fill the gap
	gxFloodFill(x+5+1, y-5-border+1, border_colour, COLOR_None);

	// fill the corners
	if (fill_colour != COLOR_None) {
		gxCircleQuadrant(x+5, y+5, 10, fill_colour, 1, TRUE);
		gxCircleQuadrant(x+w-5, y+5, 10, fill_colour, 2, TRUE);
		gxCircleQuadrant(x+5, y+h-5, 10, fill_colour, 3, TRUE);
		gxCircleQuadrant(x+w-5, y+h-5, 10, fill_colour, 4, TRUE);
	}
#endif
}