#ifdef WIN32
#pragma comment(lib, "c:\\TopfieldDisk\\DataFiles\\SDK\\TAP_SDK_DLL")
#include "TAP_Emulations.h"
#endif

#include "tap.h"
#include "skin.h"
#include "Config.h"

static char buffer[256];
static char comment[100];
static bool footer_defined;
static bool end_of_skin;


void initialiseElements()
{
	frame_colour			= RGB8888(150,150,150);	// grey
	text_colour				= RGB8888(220,220,220);	// white
	main_info_colour		= RGB8888(2,63,126);	// blue
	sub_info_colour			= RGB8888(181,18,18);	// red
	config_mode_colour		= RGB8888(32,162,32);	// green
	saved_settings_colour	= RGB8888(255,75,15);	// orange
}

bool readSkinFile(char *skin_name)
{
	TYPE_File *skin_file;
	dword file_length;

	if (!TAP_Hdd_Exist(skin_name)) return FALSE;		
	
	skin_file = TAP_Hdd_Fopen(skin_name);
	if (skin_file == NULL) return FALSE;					

	skin_buffer = TAP_MemAlloc(8192);
	memset(skin_buffer, '\0', 8192);
	skin_pointer = 0;

	file_length = TAP_Hdd_Flen(skin_file);				
	if ((file_length > 0)&& (file_length % 512) == 0)
		file_length = file_length + 512;

	TAP_Hdd_Fread(skin_buffer, file_length, 1, skin_file);

	TAP_Hdd_Fclose(skin_file);
									
	return TRUE;
}

void loadSkinElements()
{
	initialiseElements();
	end_of_skin = FALSE;

	if (readSkinFile(skin_file_name)) {
		skin_pointer = 0;
		while (skin_pointer < 8192 && !end_of_skin) {
			loadSkinElement();
		}
		TAP_MemFree(skin_buffer);
	}
}

void loadSkinElement()
{
	int i;
	char *old_pointer;
	char line_string[257];
	char element_string[257];
	bool end_of_element = FALSE;

	memset(line_string, '\0', 257);
	memset(element_string, '\0', 257);
	old_pointer = (char *) skin_buffer + skin_pointer;

	i = 0;
	while (skin_pointer < 8192 && i < 256) {
		line_string[i] = skin_buffer[skin_pointer];					

		if (line_string[i] == '\0') {
			end_of_skin = TRUE;
			break;			
		}
		if (line_string[i] == '\n') break;								
		if (line_string[i] == '\r') {															
			skin_pointer++;
			break;
		}

		if (!end_of_element && line_string[i] == ' ' && (line_string[0] != '/' && line_string[0] != '\0')) {
			memcpy(element_string, line_string, i);
			element_string[i] = '\0';
			end_of_element = TRUE;
		}
			
		skin_pointer++;
		i++;
	}

	skin_buffer[skin_pointer] = '\0';		
	if (strlen(element_string) == 0 && strlen(line_string) > 0) {
		strcpy(element_string, line_string);
		element_string[i] = '\0';
	}
	skin_pointer++;

	if (strlen(element_string) == 0 || element_string[0] == '\0' || element_string[0] == '\r' || element_string[0] == '\n' || element_string[0] == ' ' || element_string[0] == '/')
		return;

	translateSkinElement(element_string);
}

void translateSkinElement(char *element_string)
{
	int i;
	char element_name[40], *rgb_value;

	rgb_value = (char *) strrchr(element_string, '=');

	memset(element_name, '\0', 40);
	strncat(element_name, element_string, strlen(element_string)-strlen(rgb_value));
	rgb_value++;

	for (i = 0; i < SKIN_ELEMENT_COUNT; i++) {
		if (strcmp(skin_element_names[i].name, element_name) == 0)
			extractRGB(element_name, rgb_value);
	}
}

void extractRGB(char *element_name, char *rgb_value)
{
	char r[4], *g, *b, temp[15];

	b = (char *) strrchr(rgb_value, ',');
	memset(temp, '\0', 15);
	strncat(temp, rgb_value, strlen(rgb_value)-strlen(b));
	g = (char *) strrchr(temp, ',');
	memset(r, '\0', 4);
	strncat(r, rgb_value, strlen(rgb_value)-strlen(g)-strlen(b));

	b++; g++;

	if (strcmp(element_name, "Frame") == 0)
		frame_colour = RGB8888(convertRGB(r), convertRGB(g), convertRGB(b));
	ef (strcmp(element_name, "TextForeground") == 0)
		text_colour = RGB8888(convertRGB(r), convertRGB(g), convertRGB(b));
	ef (strcmp(element_name, "MainInfoBox") == 0)
		main_info_colour = RGB8888(convertRGB(r), convertRGB(g), convertRGB(b));
	ef (strcmp(element_name, "SubInfoBox") == 0)
		sub_info_colour = RGB8888(convertRGB(r), convertRGB(g), convertRGB(b));
	ef (strcmp(element_name, "ConfigModeBox") == 0)
		config_mode_colour = RGB8888(convertRGB(r), convertRGB(g), convertRGB(b));
	ef (strcmp(element_name, "SavedSettingsBox") == 0)
		saved_settings_colour = RGB8888(convertRGB(r), convertRGB(g), convertRGB(b));
}

word convertRGB(char *value)
{
    char 	str[5];
	word	tmp;
	int i;
	
	strcpy(str, value);
	tmp = 0;
	i = 0;

	while (str[i] != '\0') {
	    tmp = (tmp * 10) + (str[i] - '0');
		i++;
	}

	return tmp;
}

void writeSkinFromMemory(char *skin_name, int number_of_elements)
{
	TYPE_File *skin_file;
	word i;

	byte_count = 0;

	if (TAP_Hdd_Exist(skin_name)) 
		TAP_Hdd_Delete(skin_name);

	TAP_Hdd_Create(skin_name, ATTR_PROGRAM);					

	skin_file = TAP_Hdd_Fopen(skin_name);
	if (skin_file == NULL) 
		return; 									

	i = 0;
	memset(buffer, '\0', 512);
	while ((byte_count % 512) != 510)
	{
		strcpy(&buffer[i], " ");
		byte_count = byte_count + 1;	
		i++;
	}
	strcpy(&buffer[i], "\r");
	strcpy(&buffer[i+1], "\n");
	TAP_Hdd_Fwrite((void*) buffer, (dword) strlen(buffer), 1, skin_file);

	TAP_Hdd_Fclose(skin_file);
}

void checkForDefaultSkin()
{
	if (TAP_Hdd_Exist(DEFAULT_SKIN_NAME)) 
		return;

	writeSkinFromMemory(DEFAULT_SKIN_NAME, 0);
}